#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener Código de Trazabilidad de Granos
del web service WSCTG versión 3.0 de AFIP (RG3593/14)
"""

import os, sys, time, base64
from php import date
import traceback
from py3simplesoap.client import SoapFault
import utils

# importo funciones compartidas:
from .utils import (
    leer,
    escribir,
    leer_dbf,
    guardar_dbf,
    N,
    A,
    I,
    json,
    BaseWS,
    inicializar_y_capturar_excepciones,
    get_install_dir
)


__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010-2014 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.13b"

LICENCIA = """
wsctgv2.py: Interfaz para generar Código de Trazabilidad de Granos AFIP v1.1
Copyright (C) 2014-2015 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/CodigoTrazabilidadGranos

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA="""
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --formato: muestra el formato de los archivos de entrada/salida
  --prueba: genera y autoriza una CTG de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)

  --dummy: consulta estado de servidores
  --solicitar: obtiene el CTG (según archivo de entrada en TXT o CSV)
  --confirmar: confirma el CTG (según archivo de entrada en TXT o CSV)
  --anular: anula el CTG
  --rechazar: permite al destino rechazar el CTG
  --confirmar_arribo: confirma el arribo de un CTG
  --confirmar_definitivo: confirma el arribo definitivo de un CTG
  --regresar_a_origen_rechazado: tomar la acción de "Regresar a Origen"
  --cambiar_destino_destinatario_rechazado: "Cambio de Destino y Destinatario"

  --consultar: consulta las CTG generadas
  --consultar_excel: consulta las CTG generadas (genera un excel)
  --consultar_detalle: obtiene el detalle de una CTG
  --consultar_constancia_pdf: descarga el documento PDF de una CTG
  --pendientes: consulta CTGs otorgados, rechazados, confirmados a resolver
  --consultar_rechazados: obtener CTGs rechazados para darles un nuevo curso
  --consultar_activos_por_patente: consulta de CTGs activos por patente

  --provincias: obtiene el listado de provincias
  --localidades: obtiene el listado de localidades por provincia
  --especies: obtiene el listado de especies
  --cosechas: obtiene el listado de cosechas

Ver wsctg.ini para parámetros de configuración (URL, certificados, etc.)"
"""


# constantes de configuración (homologación):

WSDL = "https://fwshomo.afip.gov.ar/wsctg/services/CTGService_v3.0?wsdl"

DEBUG = False
XML = False
CONFIG_FILE = "wsctg.ini"
HOMO = False

# definición del formato del archivo de intercambio:

ENCABEZADO = [
    # datos enviados
    ('tipo_reg', 1, A), # 0: encabezado
    ('numero_carta_de_porte', 13, N),
    ('codigo_especie', 5, N),
    ('cuit_canjeador', 11, N),
    ('cuit_destino', 11, N),
    ('cuit_destinatario', 11, N),
    ('codigo_localidad_origen', 6, N),
    ('codigo_localidad_destino', 6, N),
    ('codigo_cosecha', 4, N),
    ('peso_neto_carga', 5, N),
    ('cant_horas', 2, N),
    ('patente_vehiculo', 6, A),
    ('cuit_transportista', 11, N),
    ('km_a_recorrer', 4, N),                          # km_recorridos (en consulta WSCTGv2)
    ('establecimiento', 6, N),                        # confirmar arribo
    ('remitente_comercial_como_canjeador', 1, A),     # S/N solicitar CTG inicial (WSCTGv2)
    ('consumo_propio', 1, A),                         # S/N confirmar arribo (WSCTGv2)

    # datos devueltos
    ('numero_ctg', 8, N),
    ('fecha_hora', 19, A),
    ('vigencia_desde', 10, A),
    ('vigencia_hasta', 10, A),
    ('transaccion', 12, N),
    ('tarifa_referencia', 6, I, 2),           # consultar detalle
    ('estado', 20, A),
    ('imprime_constancia', 5, A),
    ('observaciones', 200, A),
    ('errores', 1000, A),
    ('controles', 1000, A),
    ('detalle', 1000, A),                     # consultar detalle (WSCTGv2)

    # nuevos campos agregados:
    ('cuit_chofer', 11, N),

    # nuevos campos agregados WSCTGv3:
    ('cuit_corredor', 12, N),
    ('remitente_comercial_como_productor', 1, A),

    ]



class WSCTGv2(BaseWS):
    "Interfaz para el WebService de Código de Trazabilidad de Granos (Version 3)"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'SolicitarCTGInicial', 'SolicitarCTGDatoPendiente',
                        'ConfirmarArribo', 'ConfirmarDefinitivo',
                        'AnularCTG', 'RechazarCTG', 'CTGsPendientesResolucion',
                        'ConsultarCTG', 'LeerDatosCTG', 'ConsultarDetalleCTG',
                        'ConsultarCTGExcel', 'ConsultarConstanciaCTGPDF',
                        'ConsultarCTGRechazados',
                        'RegresarAOrigenCTGRechazado',
                        'CambiarDestinoDestinatarioCTGRechazado',
                        'ConsultarCTGActivosPorPatente',
                        'ConsultarProvincias',
                        'ConsultarLocalidadesPorProvincia',
                        'ConsultarEstablecimientos',
                        'ConsultarCosechas',
                        'ConsultarEspecies',
                        'SetParametros', 'SetParametro', 'GetParametro',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'Excepcion', 'ErrCode', 'ErrMsg', 'LanzarExcepciones', 'Errores',
        'XmlRequest', 'XmlResponse', 'Version', 'Traceback',
        'NumeroCTG', 'CartaPorte', 'FechaHora', 'CodigoOperacion',
        'CodigoTransaccion', 'Observaciones', 'Controles', 'DatosCTG',
        'VigenciaHasta', 'VigenciaDesde', 'Estado', 'ImprimeConstancia',
        'TarifaReferencia', 'Destino', 'Destinatario', 'Detalle',
        'Patente', 'PesoNeto', 'FechaVencimiento',
        'UsuarioSolicitante', 'UsuarioReal',
        ]
    _reg_progid_ = "WSCTGv2"
    _reg_clsid_ = "{ACDEFB8A-34E1-48CF-94E8-6AF6ADA0717A}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def Conectar(self, *args, **kwargs):
        ret = BaseWS.Conectar(self, *args, **kwargs)
        # corregir descripción de servicio WSDL publicado por AFIP
        # kmARecorrer -> kmRecorridos (ConsultarDetalleCTG)
        port = self.client.services['CTGService_v3.0']['ports']['CTGServiceHttpSoap20Endpoint']
        msg = port['operations']['consultarDetalleCTG']['output']['consultarDetalleCTGResponse']
        msg['response']['consultarDetalleCTGDatos']['kmRecorridos'] = int
        return ret

    def inicializar(self):
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.CodError = self.DescError = ''
        self.NumeroCTG = self.CartaPorte = ""
        self.CodigoTransaccion = self.Observaciones = ''
        self.FechaHora = self.CodigoOperacion = ""
        self.VigenciaDesde = self.VigenciaHasta = ""
        self.Controles = []
        self.DatosCTG = self.TarifaReferencia = None
        self.CodigoTransaccion = self.Observaciones = ''
        self.Detalle = self.Destino = self.Destinatario = ''
        self.Patente = self.PesoNeto = self.FechaVencimiento = ''
        self.UsuarioSolicitante = self.UsuarioReal = ''


    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'arrayErrores' in ret:
            errores = ret['arrayErrores'] or []
            self.Errores = [err['error'] for err in errores]
            self.ErrCode = ' '.join(self.Errores)
            self.ErrMsg = '\n'.join(self.Errores)

    def __analizar_controles(self, ret):
        "Comprueba y extrae controles si existen en la respuesta XML"
        if 'arrayControles' in ret:
            controles = ret['arrayControles']
            self.Controles = ["%(tipo)s: %(descripcion)s" % ctl['control']
                                for ctl in controles]

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['response']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    @inicializar_y_capturar_excepciones
    def AnularCTG(self, carta_porte, ctg):
        "Anular el CTG si se creó el mismo por error"
        response = self.client.anularCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosAnularCTG={
                            'cartaPorte': carta_porte,
                            'ctg': ctg, }))['response']
        datos = response.get('datosResponse')
        self.__analizar_errores(response)
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['CTG'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoOperacion = str(datos['codigoOperacion'])

    @inicializar_y_capturar_excepciones
    def RechazarCTG(self, carta_porte, ctg, motivo):
        "El Destino puede rechazar el CTG a través de la siguiente operatoria"
        response = self.client.rechazarCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosRechazarCTG={
                            'cartaPorte': carta_porte,
                            'ctg': ctg, 'motivoRechazo': motivo,
                            }))['response']
        datos = response.get('datosResponse')
        self.__analizar_errores(response)
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['CTG'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoOperacion = str(datos['codigoOperacion'])

    @inicializar_y_capturar_excepciones
    def SolicitarCTGInicial(self, numero_carta_de_porte, codigo_especie,
        cuit_canjeador, cuit_destino, cuit_destinatario, codigo_localidad_origen,
        codigo_localidad_destino, codigo_cosecha, peso_neto_carga,
        cant_horas=None, patente_vehiculo=None, cuit_transportista=None,
        km_a_recorrer=None, remitente_comercial_como_canjeador=None,
        cuit_corredor=None, remitente_comercial_como_productor=None,
        **kwargs):
        "Solicitar CTG Desde el Inicio"
        # ajusto parámetros según validaciones de AFIP:
        if cuit_canjeador and int(cuit_canjeador) == 0:
            cuit_canjeador = None         # nulo
        if not remitente_comercial_como_canjeador:
            remitente_comercial_como_canjeador = None

        ret = self.client.solicitarCTGInicial(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                         datosSolicitarCTGInicial=dict(
                            cartaPorte=numero_carta_de_porte,
                            codigoEspecie=codigo_especie,
                            cuitCanjeador=cuit_canjeador or None,
                            remitenteComercialComoCanjeador=remitente_comercial_como_canjeador,
                            cuitDestino=cuit_destino,
                            cuitDestinatario=cuit_destinatario,
                            codigoLocalidadOrigen=codigo_localidad_origen,
                            codigoLocalidadDestino=codigo_localidad_destino,
                            codigoCosecha=codigo_cosecha,
                            pesoNeto=peso_neto_carga,
                            cuitTransportista=cuit_transportista,
                            cantHoras=cant_horas,
                            patente=patente_vehiculo,
                            kmARecorrer=km_a_recorrer,
                            cuitCorredor=cuit_corredor,
                            remitenteComercialcomoProductor=remitente_comercial_como_productor,
                            )))['response']
        self.__analizar_errores(ret)
        self.Observaciones = ret['observacion']
        datos = ret.get('datosSolicitarCTGResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            datos_ctg = datos.get('datosSolicitarCTG')
            if datos_ctg:
                self.NumeroCTG = str(datos_ctg['ctg'])
                self.FechaHora = str(datos_ctg['fechaEmision'])
                self.VigenciaDesde = str(datos_ctg['fechaVigenciaDesde'])
                self.VigenciaHasta = str(datos_ctg['fechaVigenciaHasta'])
                self.TarifaReferencia = str(datos_ctg.get('tarifaReferencia'))
            self.__analizar_controles(datos)
        return self.NumeroCTG or 0

    @inicializar_y_capturar_excepciones
    def SolicitarCTGDatoPendiente(self, numero_carta_de_porte, cant_horas,
        patente_vehiculo, cuit_transportista):
        "Solicitud que permite completar los datos faltantes de un Pre-CTG "
        "generado anteriormente a través de la operación solicitarCTGInicial"
        ret = self.client.solicitarCTGDatoPendiente(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosSolicitarCTGDatoPendiente=dict(
                            cartaPorte=numero_carta_de_porte,
                            cuitTransportista=cuit_transportista,
                            cantHoras=cant_horas,
                            )))['response']
        self.__analizar_errores(ret)
        self.Observaciones = ret['observacion']
        datos = ret.get('datosSolicitarCTGResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            datos_ctg = datos.get('datosSolicitarCTG')
            if datos_ctg:
                self.NumeroCTG = str(datos_ctg['ctg'])
                self.FechaHora = str(datos_ctg['fechaEmision'])
                self.VigenciaDesde = str(datos_ctg['fechaVigenciaDesde'])
                self.VigenciaHasta = str(datos_ctg['fechaVigenciaHasta'])
                self.TarifaReferencia = str(datos_ctg.get('tarifaReferencia'))
            self.__analizar_controles(datos)
        return self.NumeroCTG

    @inicializar_y_capturar_excepciones
    def ConfirmarArribo(self, numero_carta_de_porte, numero_ctg,
                        cuit_transportista, peso_neto_carga,
                        consumo_propio, establecimiento=None, cuit_chofer=None,
                        **kwargs):
        "Confirma arribo CTG"
        ret = self.client.confirmarArribo(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosConfirmarArribo=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            cuitTransportista=cuit_transportista,
                            cuitChofer=cuit_chofer,
                            cantKilosCartaPorte=peso_neto_carga,
                            consumoPropio=consumo_propio,
                            establecimiento=establecimiento,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def ConfirmarDefinitivo(self, numero_carta_de_porte, numero_ctg,
                            establecimiento=None, codigo_cosecha=None, peso_neto_carga=None,
                            **kwargs):
        "Confirma arribo definitivo CTG"
        ret = self.client.confirmarDefinitivo(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosConfirmarDefinitivo=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            establecimiento=establecimiento,
                            codigoCosecha=codigo_cosecha,
                            pesoNeto=peso_neto_carga,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos.get('codigoOperacion', ""))
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def RegresarAOrigenCTGRechazado(self, numero_carta_de_porte, numero_ctg,
                            km_a_recorrer=None,
                            **kwargs):
        "Al consultar los CTGs rechazados se puede Regresar a Origen"
        ret = self.client.regresarAOrigenCTGRechazado(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosRegresarAOrigenCTGRechazado=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg, kmARecorrer=km_a_recorrer,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def CambiarDestinoDestinatarioCTGRechazado(self, numero_carta_de_porte,
                            numero_ctg, codigo_localidad_destino=None,
                            cuit_destino=None, cuit_destinatario=None,
                            km_a_recorrer=None,
                            **kwargs):
        "Tomar acción de Cambio de Destino y Destinatario para CTG rechazado"
        ret = self.client.cambiarDestinoDestinatarioCTGRechazado(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosCambiarDestinoDestinatarioCTGRechazado=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            codigoLocalidadDestino=codigo_localidad_destino,
                            cuitDestino=cuit_destino,
                            cuitDestinatario=cuit_destinatario,
                            kmARecorrer=km_a_recorrer,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def ConsultarCTG(self, numero_carta_de_porte=None, numero_ctg=None,
                     patente=None, cuit_solicitante=None, cuit_destino=None,
                     fecha_emision_desde=None, fecha_emision_hasta=None):
        "Operación que realiza consulta de CTGs según el criterio ingresado."
        ret = self.client.consultarCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        consultarCTGDatos=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            patente=patente,
                            cuitSolicitante=cuit_solicitante,
                            cuitDestino=cuit_destino,
                            fechaEmisionDesde=fecha_emision_desde,
                            fechaEmisionHasta=fecha_emision_hasta,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayDatosConsultarCTG')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return ''

    @inicializar_y_capturar_excepciones
    def ConsultarCTGRechazados(self):
        "Consulta de CTGs Otorgados, CTGs Rechazados y CTGs Confirmados"
        ret = self.client.consultarCTGRechazados(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayConsultarCTGRechazados')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return False

    @inicializar_y_capturar_excepciones
    def ConsultarCTGActivosPorPatente(self, patente="ZZZ999"):
        "Consulta de CTGs activos por patente"
        ret = self.client.consultarCTGActivosPorPatente(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        patente=patente,
                            ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayConsultarCTGActivosPorPatenteResponse')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return False

    @inicializar_y_capturar_excepciones
    def CTGsPendientesResolucion(self):
        "Consulta de CTGs Otorgados, CTGs Rechazados y CTGs Confirmados"
        ret = self.client.CTGsPendientesResolucion(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        if ret:
            self.DatosCTG = ret
            return True
        else:
            self.DatosCTG = {}
        return False

    def LeerDatosCTG(self, clave='', pop=True):
        "Recorro los datos devueltos y devuelvo el primero si existe"

        if clave and self.DatosCTG:
            # obtengo la lista por estado pendiente de resolución ("array")
            datos = self.DatosCTG[clave]
        else:
            # uso directamente la lista devuelta por la consulta
            datos = self.DatosCTG
        if datos:
            # extraigo el primer item
            if pop:
                datos = datos.pop(0)
            else:
                datos = datos[0]
            for det in ('datosConsultarCTG', 'detalleConsultaCTGRechazado',
                          'detalleConsultaCTGActivo'):
                if det in datos:
                    datos_ctg = datos[det]
                    break
            else:
                # elemento del array no encontrado:
                return ""
            self.CartaPorte = str(datos_ctg['cartaPorte'])
            self.NumeroCTG = str(datos_ctg['ctg'])
            self.Estado = unicode(datos_ctg.get('estado', ""))
            self.ImprimeConstancia = str(datos_ctg.get('imprimeConstancia', ""))
            for campo in ("fechaRechazo", "fechaEmision", "fechaSolicitud",
                          "fechaConfirmacionArribo"):
                if campo in datos_ctg:
                    self.FechaHora = str(datos_ctg.get(campo))
            self.Destino = datos_ctg.get("destino", "")
            self.Destinatario = datos_ctg.get("destinatario", "")
            self.Observaciones = datos_ctg.get("observaciones", "")
            self.Patente = datos_ctg.get("patente")
            self.PesoNeto = datos_ctg.get("pesoNeto")
            self.FechaVencimiento = datos_ctg.get("fechaVencimiento")
            self.UsuarioSolicitante = datos_ctg.get("usuarioSolicitante")
            self.UsuarioReal = datos_ctg.get("usuarioReal")
            return self.NumeroCTG
        else:
            return ""

    @inicializar_y_capturar_excepciones
    def ConsultarCTGExcel(self, numero_carta_de_porte=None, numero_ctg=None,
                     patente=None, cuit_solicitante=None, cuit_destino=None,
                     fecha_emision_desde=None, fecha_emision_hasta=None,
                     archivo="planilla.xls"):
        "Operación que realiza consulta de CTGs, graba una planilla xls"
        ret = self.client.consultarCTGExcel(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        consultarCTGDatos=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            patente=patente,
                            cuitSolicitante=cuit_solicitante,
                            cuitDestino=cuit_destino,
                            fechaEmisionDesde=fecha_emision_desde,
                            fechaEmisionHasta=fecha_emision_hasta,
                            )))['response']
        self.__analizar_errores(ret)
        datos = base64.b64decode(ret.get('archivo') or "")
        f = open(archivo, "wb")
        f.write(datos)
        f.close()
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarDetalleCTG(self, numero_ctg=None):
        "Operación mostrar este detalle de la  solicitud de CTG seleccionada."
        ret = self.client.consultarDetalleCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ctg=numero_ctg,
                        ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('consultarDetalleCTGDatos')
        if datos:
            self.NumeroCTG = str(datos['ctg'])
            self.CartaPorte = str(datos['cartaPorte'])
            self.Estado = unicode(datos['estado'])
            self.FechaHora = str(datos['fechaEmision'])
            self.VigenciaDesde = str(datos['fechaVigenciaDesde'])
            self.VigenciaHasta = str(datos['fechaVigenciaHasta'])
            self.TarifaReferencia = str(datos['tarifaReferencia'])
            self.Detalle = str(datos.get('detalle', ""))
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarConstanciaCTGPDF(self, numero_ctg=None,
                                        archivo="constancia.pdf"):
        "Operación Consultar Constancia de CTG en PDF"
        ret = self.client.consultarConstanciaCTGPDF(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ctg=numero_ctg,
                        ))['response']
        self.__analizar_errores(ret)
        datos = base64.b64decode(ret.get('archivo', ""))
        f = open(archivo, "wb")
        f.write(datos)
        f.close()
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarProvincias(self, sep="||"):
        ret = self.client.consultarProvincias(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['consultarProvinciasResponse']
        self.__analizar_errores(ret)
        array = ret.get('arrayProvincias', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['provincia']['codigo'],
                     it['provincia']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarLocalidadesPorProvincia(self, codigo_provincia, sep="||"):
        ret = self.client.consultarLocalidadesPorProvincia(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        codigoProvincia=codigo_provincia,
                        ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayLocalidades', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['localidad']['codigo'],
                     it['localidad']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarEstablecimientos(self, sep="||"):
        ret = self.client.consultarEstablecimientos(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayEstablecimientos', [])
        return [("%s" %
                    (it['establecimiento'],))
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarEspecies(self, sep="||"):
        ret = self.client.consultarEspecies(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayEspecies', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['especie']['codigo'],
                     it['especie']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarCosechas(self, sep="||"):
        ret = self.client.consultarCosechas(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayCosechas', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['cosecha']['codigo'],
                     it['cosecha']['descripcion'])
               for it in array]


def leer_archivo(nombre_archivo):
    archivo = open(nombre_archivo, "r")
    items = []
    ext = os.path.splitext(nombre_archivo)[1]
    if ext == '.csv':
        csv_reader = csv.reader(open(ENTRADA), dialect='excel', delimiter=";")
        for row in csv_reader:
            items.append(row)
        cols = [str(it).strip() for it in items[0]]
        # armar diccionario por cada linea
        items = [dict([(cols[i],str(v).strip()) for i,v in enumerate(item)]) for item in items[1:]]
        return cols, items
    elif ext == '.json':
        items = json.load(archivo)
    elif ext == '.dbf':
        dic = {}
        formatos = [('Encabezado', ENCABEZADO, dic), ]
        leer_dbf(formatos, conf_dbf)
        items = [dic]
    elif ext == '.txt':
        dic = {}
        for linea in archivo:
            if str(linea[0])=='0':
                dic.update(leer(linea, ENCABEZADO))
            else:
                print("Tipo de registro incorrecto:", linea[0])
        items.append(dic)
    else:
        raise RuntimeError("Extension de archivo desconocida: %s" % ext)
    archivo.close()
    cols = [k[0] for k in ENCABEZADO]
    return cols, items


def escribir_archivo(cols, items, nombre_archivo, agrega=False):
    archivo = open(nombre_archivo, agrega and "a" or "w")
    ext = os.path.splitext(nombre_archivo)[1]
    if ext == '.csv':
        csv_writer = csv.writer(archivo, dialect='excel', delimiter=";")
        csv_writer.writerows([cols])
        csv_writer.writerows([[item[k] for k in cols] for item in items])
    elif ext == '.json':
        json.dump(items, archivo, sort_keys=True, indent=4)
    elif ext == '.dbf':
        formatos = [('Encabezado', ENCABEZADO, items), ]
        guardar_dbf(formatos, True, conf_dbf)
    elif ext == '.txt':
        for dic in items:
            dic['tipo_reg'] = 0
            archivo.write(escribir(dic, ENCABEZADO))
    else:
        raise RuntimeError("Extension de archivo desconocida: %s" % ext)
    archivo.close()



# busco el directorio de instalación (global para que no cambie si usan otra dll)
if not hasattr(sys, "frozen"):
    basepath = __file__
elif sys.frozen=='dll':
    import win32api
    basepath = win32api.GetModuleFileName(sys.frozendllhandle)
else:
    basepath = sys.executable
INSTALL_DIR = WSCTGv2.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)
    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato in [('Encabezado', ENCABEZADO), ]:
            comienzo = 1
            print("=== %s ===" % msg)
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                dec = len(fmt)>3 and fmt[3] or (tipo=='I' and '2' or '')
                print(" * Campo: %-20s Posición: %3d Longitud: %4d Tipo: %s Decimales: %s" % (
                    clave, comienzo, longitud, tipo, dec))
                comienzo += longitud
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSCTGv2)
        sys.exit(0)

    import csv
    from ConfigParser import SafeConfigParser

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        for arg in sys.argv[1:]:
            if arg.startswith("--"):
                break
            print("Usando configuración:", arg)
            CONFIG_FILE = arg

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA','CERT')
        PRIVATEKEY = config.get('WSAA','PRIVATEKEY')
        CUIT = config.get('WSCTG','CUIT')
        ENTRADA = config.get('WSCTG','ENTRADA')
        SALIDA = config.get('WSCTG','SALIDA')

        if config.has_option('WSAA','URL') and not HOMO:
            wsaa_url = config.get('WSAA','URL')
        else:
            wsaa_url = None
        if config.has_option('WSCTG','URL') and not HOMO:
            wsctg_url = config.get('WSCTG','URL')
        else:
            wsctg_url = WSDL

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG:
                print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("wsaa_url:", wsaa_url)
            print("wsctg_url:", wsctg_url)

        # obteniendo el TA
        from wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsctg", CERT, PRIVATEKEY, wsaa_url, debug=DEBUG)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wsctg = WSCTGv2()
        wsctg.Conectar(wsdl=wsctg_url)
        wsctg.SetTicketAcceso(ta)
        wsctg.Cuit = CUIT

        if '--dummy' in sys.argv:
            ret = wsctg.Dummy()
            print("AppServerStatus", wsctg.AppServerStatus)
            print("DbServerStatus", wsctg.DbServerStatus)
            print("AuthServerStatus", wsctg.AuthServerStatus)
            sys.exit(0)

        if '--anular' in sys.argv:
            i = sys.argv.index("--anular")
            # print(wsctg.client.help("anularCTG"))
            if i + 2 > len(sys.argv) or sys.argv[i + 1].startswith("--"):
                carta_porte = raw_input("Ingrese Carta de Porte: ")
                ctg = raw_input("Ingrese CTG: ")
            else:
                carta_porte = sys.argv[i + 1]
                ctg = sys.argv[i + 2]
            ret = wsctg.AnularCTG(carta_porte, ctg)
            wsctg.SolicitarCTGDatoPendiente()
            print("Carta Porte", wsctg.CartaPorte)
            print("Numero CTG", wsctg.NumeroCTG)
            print("Fecha y Hora", wsctg.FechaHora)
            print("Codigo Anulacion de CTG", wsctg.CodigoOperacion)
            print("Errores:", wsctg.Errores)
            sys.exit(0)

        if '--rechazar' in sys.argv:
            i = sys.argv.index("--rechazar")
            # print(wsctg.client.help("rechazarCTG"))
            if i + 3 > len(sys.argv) or sys.argv[i + 1].startswith("--"):
                carta_porte = raw_input("Ingrese Carta de Porte: ")
                ctg = raw_input("Ingrese CTG: ")
                motivo = raw_input("Motivo: ")
            else:
                carta_porte = sys.argv[i + 1]
                ctg = sys.argv[i + 2]
                motivo = sys.argv[i + 3]
            ret = wsctg.RechazarCTG(carta_porte, ctg, motivo)
            print("Carta Porte", wsctg.CartaPorte)
            print("Numero CTG", wsctg.NumeroCTG)
            print("Fecha y Hora", wsctg.FechaHora)
            print("Codigo Anulacion de CTG", wsctg.CodigoOperacion)
            print("Errores:", wsctg.Errores)
            sys.exit(0)


        # Recuperar parámetros:

        if '--provincias' in sys.argv:
            ret = wsctg.ConsultarProvincias()
            print("\n".join(ret))

        if '--localidades' in sys.argv:
            ret = wsctg.ConsultarLocalidadesPorProvincia(16)
            print("\n".join(ret))

        if '--especies' in sys.argv:
            ret = wsctg.ConsultarEspecies()
            print("\n".join(ret))

        if '--cosechas' in sys.argv:
            ret = wsctg.ConsultarCosechas()
            print("\n".join(ret))

        if '--establecimientos' in sys.argv:
            ret = wsctg.ConsultarEstablecimientos()
            print("\n".join(ret))


        if '--prueba' in sys.argv or '--formato' in sys.argv:
            prueba = dict(numero_carta_de_porte=512345679, codigo_especie=23,
                cuit_canjeador=0, #30660685908,
                cuit_destino=20111111112, cuit_destinatario=20222222223,
                codigo_localidad_origen=3058, codigo_localidad_destino=3059,
                codigo_cosecha='1314', peso_neto_carga=1000,
                km_a_recorrer=1234,
                observaciones='', establecimiento=1,
            )
            if [argv for argv in sys.argv if argv.startswith(("--confirmar",
                    "--regresar", '--cambiar'))]:
                prueba.update(dict(
                    numero_ctg="49241727", transaccion='10000001681',
                    consumo_propio='S',
                    ))
            parcial = dict(
                    cant_horas=1,
                    patente_vehiculo='APE652', cuit_transportista=20333333334,
                    )
            if not '--parcial' in sys.argv:
                prueba.update(parcial)

            escribir_archivo(prueba.keys(), [prueba], ENTRADA)

        cols, items = leer_archivo(ENTRADA)
        ctg = None

        if '--solicitar' in sys.argv:
            wsctg.LanzarExcepciones = True
            for it in items:
                print("solicitando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                ctg = wsctg.SolicitarCTGInicial(**it)
                print("numero CTG: ", ctg)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['tarifa_referencia'] = wsctg.TarifaReferencia
                it['observaciones'] = wsctg.Observaciones
                it['fecha_hora'] = wsctg.FechaHora
                it['vigencia_desde'] = wsctg.VigenciaDesde
                it['vigencia_hasta'] = wsctg.VigenciaHasta
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--parcial' in sys.argv:
            wsctg.LanzarExcepciones = True
            for it in items:
                print("solicitando dato pendiente...", ' '.join(['%s=%s' % (k,v) for k,v in parcial.items()]))
                ctg = wsctg.SolicitarCTGDatoPendiente(
                    numero_carta_de_porte=wsctg.CartaPorte,
                    **parcial)
                print("numero CTG: ", ctg)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['tarifa_referencia'] = wsctg.TarifaReferencia
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--confirmar_arribo' in sys.argv:
            for it in items:
                print("confirmando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.ConfirmarArribo(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--confirmar_definitivo' in sys.argv:
            if '--testing' in sys.argv:
                wsctg.LoadTestXML("wsctg_confirmar_def.xml") # cargo respuesta
            for it in items:
                print("confirmando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.ConfirmarDefinitivo(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--regresar_a_origen_rechazado' in sys.argv:
            for it in items:
                print("regresando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.RegresarAOrigenCTGRechazado(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--cambiar_destino_destinatario_rechazado' in sys.argv:
            for it in items:
                print("cambiando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.CambiarDestinoDestinatarioCTGRechazado(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--consultar_detalle' in sys.argv:
            i = sys.argv.index("--consultar_detalle")
            if len(sys.argv) > i + 1 and not sys.argv[i+1].startswith("--"):
                ctg = int(sys.argv[i+1])
            elif not ctg:
                ctg = int(raw_input("Numero de CTG: ") or '0') or 73714620

            wsctg.LanzarExcepciones = True
            for i, it in enumerate(items):
                print("consultando detalle...", ctg)
                ok = wsctg.ConsultarDetalleCTG(ctg)
                print("Numero CTG: ", wsctg.NumeroCTG)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                print("Detalle:", wsctg.Detalle)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['observaciones'] = wsctg.Observaciones
                it['fecha_hora'] = wsctg.FechaHora
                it['vigencia_desde'] = wsctg.VigenciaDesde
                it['vigencia_hasta'] = wsctg.VigenciaHasta
                wsctg.AnalizarXml("XmlResponse")
                for k, ki in {'ctg': 'numero_ctg', 'solicitante': '',
                          'estado': 'estado',
                          'especie': '', ##'codigo_especie', no devuelve codigo!
                          'cosecha': '', ##'codigo_cosecha', no devuelve codigo!
                          'cuitCanjeador': 'cuit_canjeador',
                          'cuitDestino': 'cuit_destino',
                          'cuitDestinatario': 'cuit_destinatario',
                          'cuitTransportista': 'cuit_transportista',
                          'establecimiento': 'establecimiento',
                          'localidadOrigen': 'localidad_origen',
                          'localidadDestino': 'localidad_destino',
                          'cantidadHoras': 'cantidad_horas',
                          'patenteVehiculo': 'patente_vehiculo',
                          'pesoNetoCarga': 'peso_neto_carga',
                          'kmRecorridos': 'km_recorridos',
                          'tarifaReferencia': 'tarifa_referencia', }.items():
                    v = wsctg.ObtenerTagXml('consultarDetalleCTGDatos', k)
                    print(k, v)
                    if ki.startswith("cuit") and v:
                        v = v[:11]
                    it[ki] = v

        escribir_archivo(cols, items, SALIDA)

        if "--consultar" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.ConsultarCTG(fecha_emision_desde="01/04/2012")
            print("Numero CTG - Carta de Porte - Imprime Constancia - Estado")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte,)
                print(wsctg.ImprimeConstancia, wsctg.Estado, wsctg.FechaHora)

        if "--consultar_rechazados" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.ConsultarCTGRechazados()
            print("Numero CTG - Carta de Porte - Fecha - Destino/Dest./Obs.")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.FechaHora,)
                print(wsctg.Destino, wsctg.Destinatario, wstcg.Observaciones)

        if "--consultar_activos_por_patente" in sys.argv:
            i = sys.argv.index("--consultar_activos_por_patente")
            if len(sys.argv) > i + 1 and not sys.argv[i+1].startswith("--"):
                patente = int(sys.argv[i+1])
            elif not ctg:
                patente= raw_input("Patente: ") or 'APE652'
            wsctg.LanzarExcepciones = True
            if '--testing' in sys.argv:
                wsctg.LoadTestXML("wsctgv2_activos.xml")
            wsctg.ConsultarCTGActivosPorPatente(patente=patente)
            print("Numero CTG - Carta de Porte - Fecha - Peso Neto - Usuario")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.Patente,)
                print(wsctg.FechaHora, wsctg.FechaVencimiento, wsctg.PesoNeto,)
                print(wsctg.UsuarioSolicitante, wsctg.UsuarioReal)

        if '--consultar_excel' in sys.argv:
            archivo = raw_input("Archivo a generar (planilla.xls): ") or \
                        'planilla.xls'
            wsctg.LanzarExcepciones = True
            ok = wsctg.ConsultarCTGExcel(fecha_emision_desde="01/04/2012",
                                         archivo=archivo)
            print("Errores:", wsctg.Errores)

        if '--consultar_constancia_pdf' in sys.argv:
            i = sys.argv.index("--consultar_constancia_pdf")
            if len(sys.argv) > i + 2 and not sys.argv[i+1].startswith("--"):
                ctg = int(sys.argv[i+1])
                archivo = sys.argv[i+2]
            elif not ctg:
                ctg = int(raw_input("Numero de CTG: ") or '0') or 83139794
                archivo = raw_input("Archivo a generar (constancia.pdf): ") or \
                            'constancia.pdf'

            wsctg.LanzarExcepciones = True
            ok = wsctg.ConsultarConstanciaCTGPDF(ctg, archivo)
            print("Errores:", wsctg.Errores)

        if "--pendientes" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.CTGsPendientesResolucion()
            for clave in ("arrayCTGsRechazadosAResolver",
                          "arrayCTGsOtorgadosAResolver",
                          "arrayCTGsConfirmadosAResolver", ):
                print(clave[6:])
                print("Numero CTG - Carta de Porte - Imprime Constancia - Estado")
                while wsctg.LeerDatosCTG(clave):
                    print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.FechaHora)
                    print(wsctg.Destino, wsctg.Destinatario, wsctg.Observaciones)

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii","ignore"))
        sys.exit(3)
    except Exception as e:
        ex = utils.exception_info()
        print(ex)
        if DEBUG:
            raise
        sys.exit(5)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener código de autorización de impresión o
electrónico del web service WSBFE de AFIP (Bono Fiscal Electrónico)
"""

__author__ = "Mariano Reingart <mariano@nsis.com.ar>"
__copyright__ = "Copyright (C) 2009 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.14"

import sys
from php import date, SimpleXMLElement, SoapClient

WSBFEURL = "https://wswhomo.afip.gov.ar/wsbfe/service.asmx"
##WSBFEURL = "https://servicios1.afip.gov.ar/wsbfe/service.asmx"
SOAP_ACTION = 'http://ar.gov.afip.dif.bfe/'
SOAP_NS = "http://ar.gov.afip.dif.bfe/"

class BFEError(RuntimeError):
    "Clase para informar errores del WSBFE"
    def __init__(self, bfeerror):
        self.code = str(bfeerror.ErrCode)
        self.msg = unicode(bfeerror.ErrMsg)

    def __unicode__(self):
        return u"BFEError %s: %s" % (self.code, self.msg)

def get_param_mon(client, token, sign, cuit):
    "Recuperador de valores referenciales de codigos de Moneda"
    response = client.BFEGetPARAM_MON(
        auth={"Token": token, "Sign": sign, "Cuit":long(cuit)})

    if int(response.BFEGetPARAM_MONResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_MONResult.BFEErr)

    mons = [] # monedas
    for m in response.BFEGetPARAM_MONResult.BFEResultGet.ClsBFEResponse_Mon:
        mon = {'id': str(m.Mon_Id), 'ds': unicode(m.Mon_Ds),
               'vig_desde': str(m.Mon_vig_desde),
               'vig_hasta': str(m.Mon_vig_hasta)}
        mons.append(mon)
    return mons

def get_param_ncm(client, token, sign, cuit):
    "Recuperador de valores referenciales de códigos de productos"
    response = client.BFEGetPARAM_NCM(
        Auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.BFEGetPARAM_NCMResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_NCMResult.BFEErr)

    ncms = [] # productos
    for n in response.BFEGetPARAM_NCMResult.BFEResultGet.ClsBFEResponse_NCM:
        ncm = {'codigo': str(n.NCM_Codigo), 'ds': unicode(n.NCM_Ds),
               'nota': unicode(n.NCM_Nota),
               'vig_desde': str(n.NCM_vig_desde),
               'vig_hasta': str(n.NCM_vig_hasta)}
        ncms.append(ncm)
    return ncms

def get_param_tipo_cbte(client, token, sign, cuit):
    "Recuperador de valores referenciales de códigos de Tipos de comprobante"
    response = client.BFEGetPARAM_Tipo_Cbte(
        auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.BFEGetPARAM_Tipo_CbteResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_Tipo_CbteResult.BFEErr)

    cbtes = [] # tipos de comprobantes
    for c in response.BFEGetPARAM_Tipo_CbteResult.BFEResultGet.ClsBFEResponse_Tipo_Cbte:
        cbte = {'id': int(c.Cbte_Id), 'ds': unicode(c.Cbte_Ds),
                'vig_desde': str(c.Cbte_vig_desde),
                'vig_hasta': str(c.Cbte_vig_hasta)}
        cbtes.append(cbte)
    return cbtes

def get_param_tipo_iva(client, token, sign, cuit):
    "Recuperador de valores referenciales de códigos de IVA"
    response = client.BFEGetPARAM_Tipo_IVA(
        auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.BFEGetPARAM_Tipo_IVAResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_Tipo_IVAResult.BFEErr)

    ivas= [] # tipos de iva
    for i in response.BFEGetPARAM_Tipo_IVAResult.BFEResultGet.ClsBFEResponse_Tipo_IVA:
        iva = {'id': int(i.IVA_Id), 'ds': unicode(i.IVA_Ds),
                'vig_desde': str(i.IVA_vig_desde),
                'vig_hasta': str(i.IVA_vig_hasta)}
        ivas.append(iva)
    return ivas

def get_param_umed(client, token, sign, cuit):
    "Recuperador de valores referenciales de Unidades de Medidas"
    response = client.BFEGetPARAM_UMed(
        auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.BFEGetPARAM_UMedResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_UMedResult.BFEErr)

    umeds = [] # unidades de medida
    for u in response.BFEGetPARAM_UMedResult.BFEResultGet.ClsBFEResponse_UMed:
        umed = {'id': int(u.Umed_Id), 'ds': unicode(u.Umed_Ds),
                'vig_desde': str(u.Umed_vig_desde),
                'vig_hasta': str(u.Umed_vig_hasta)}
        umeds.append(umed)
    return umeds

def get_param_zonas(client, token, sign, cuit):
    "Recuperador de valores referenciales de Zonas"
    response = client.BFEGetPARAM_Zonas(
        auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)})

    if int(response.BFEGetPARAM_ZonasResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetPARAM_ZonasResult.BFEErr)

    zonas = [] # unidades de medida
    for z in response.BFEGetPARAM_ZonasResult.BFEResultGet.ClsBFEResponse_Zon:
        zon = {'id': int(z.Zon_Id), 'ds': unicode(z.Zon_Ds),
                'vig_desde': str(z.Zon_vig_desde),
                'vig_hasta': str(z.Zon_vig_hasta)}
        zonas.append(zon)
    return zonas

def authorize(client, token, sign, cuit, id, factura):
    "Autorizador: recibe la factura como un dict, devuelve auth y events"

    # Información de la autenticación
    auth= {"Token": token, "Sign": sign, "Cuit": long(cuit)}
    # Información de la factura de ingreso
    cmp={"Id": id,}
    cmp.update(factura)

    # llamo al webservice para autorizar la factura
    response = client.BFEAuthorize(Auth=auth, Cmp=cmp)

    # hubo error?
    if int(response.BFEAuthorizeResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEAuthorizeResult.BFEErr)

    # extraigo la respuesta (auth y eventos)
    result = response.BFEAuthorizeResult.BFEResultAuth
    auth = dict(id=int(result.Id), cuit=int(result.Cuit), cae=str(result.Cae),
               fch_cbte=str(result.Fch_cbte), resultado=str(result.Resultado),
               fch_venc_cae=str(result.Fch_venc_Cae),
               reproceso=str(result.Reproceso), obs=str(result.Obs))
    events = []
    for bfe_event in response.BFEAuthorizeResult.BFEEvents:
        events.append(dict(code=bfe_event.EventCode, msg=bfe_event.EventMsg))

    return auth, events

def get_cmp(client, token, sign, cuit, tipo_cbte, punto_vta, cbte_nro):
    "Recupera los datos completos de un comprobante ya autorizado"
    response = client.BFEGetCMP(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)},
        Cmp={"Tipo_cbte": tipo_cbte,
             "Punto_vta": punto_vta, "Cbte_nro": cbte_nro})

    if int(response.BFEGetCMPResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetCMPResult.BFEErr)

    result = response.BFEGetCMPResult.BFEResultGet
    # print(result.Id)
    cbt = dict(cuit=int(result.Cuit), cae=str(result.Cae),
               fch_cbte=str(result.Fecha_cbte_orig),
               fch_cae=str(result.Fecha_cbte_cae),
               fch_venc_cae=str(result.Fch_venc_Cae),
               imp_total=str(result.Imp_total),
               imp_neto=str(result.Imp_neto),
               impto_liq=str(result.Impto_liq),
               obs=str(result.Obs))
    events = []
    for bfe_event in response.BFEGetCMPResult.BFEEvents:
        events.append(dict(code=bfe_event.EventCode, msg=bfe_event.EventMsg))
    return cbt, events

def get_last_cmp(client, token, sign, cuit, tipo_cbte, punto_vta):
    "Recupera el ultimos comprobante autorizado"
    response = client.BFEGetLast_CMP(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit),
              "Tipo_cbte": tipo_cbte,
              "Pto_venta": punto_vta})

    if int(response.BFEGetLast_CMPResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetLast_CMPResult.BFEErr)

    result = response.BFEGetLast_CMPResult.BFEResult_LastCMP
    cbte_nro = int(result.Cbte_nro)
    try:
        cbte_fecha = str(result.Cbte_fecha)
    except AttributeError:
        cbte_fecha = ""

    events = []
    for bfe_event in response.BFEGetLast_CMPResult.BFEEvents:
        events.append(dict(code=bfe_event.EventCode, msg=bfe_event.EventMsg))
    return cbte_nro, cbte_fecha, events

def get_last_id(client, token, sign, cuit):
    "Recupera el ultimo ID y su fecha"
    response = client.BFEGetLast_ID(
        Auth={"Token": token, "Sign": sign, "Cuit": long(cuit)},
        )

    if int(response.BFEGetLast_IDResult.BFEErr.ErrCode) != 0:
        raise BFEError(response.BFEGetLast_IDResult.BFEErr)

    result = response.BFEGetLast_IDResult.BFEResultGet
    id = int(result.Id)
    events = []
    for bfe_event in response.BFEGetLast_IDResult.BFEEvents:
        events.append(dict(code=bfe_event.EventCode, msg=bfe_event.EventMsg))
    return id, events

def dummy(client):
    "Metodo dummy para verificacion de funcionamiento"
    response = client.BFEDummy()
    result = response.BFEDummyResult
    appserver = dbserver = authserver = None
    try:
        appserver = str(result.AppServer)
        dbserver = str(result.DbServer)
        authserver = str(result.AuthServer)
    except (RuntimeError, IndexError, AttributeError) as e:
        pass
    return {'appserver': appserver,
            'dbserver': dbserver,
            'authserver': authserver}

# Clases para facilitar la autorización de facturas

class FacturaBF:
    "Factura Bono Fiscal"
    # valores por defecto del encabezado de la factura
    tipo_doc = 80; nro_doc = 23111111113
    zona = 0; tipo_cbte = 1; punto_vta = 1; cbte_nro = 0; fecha_cbte = None
    imp_total = 0.0; imp_neto = 0.0; impto_liq = 0.0
    imp_tot_conc = 0.0; impto_liq_rni = 0.00; imp_op_ex = 0.00
    imp_perc = 0.00; imp_iibb = 0.00; imp_perc_mun = 0.00; imp_internos=0.00
    imp_moneda_id = 0; imp_moneda_ctz = 1.0
    # datos del cliente (para impresion)
    nombre = ""
    domicilio = ""; localidad = ""; provincia = ""
    email = ""; telefono = ""; observaciones = ""
    categoria = ""
    items = None

    def __init__(self,**kwargs):
        self.items = []
        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_item(self, item, calc=True):
        self.items.append(item)
        total = item.imp_total
        if not calc:
            return

        if item.iva_id==1: #'No gravado'
            self.imp_tot_conc += total
        elif item.iva_id==2: # exento
            self.imp_op_ex += total
        else:
            self.imp_neto += total
            iva = total * {3:0, 4:10.5, 5:21, 6:27}[item.iva_id] / 100
            self.impto_liq += iva
            total += iva
        self.imp_total += total

    def to_dict(self, completo=False):
        "Convierte el objeto factura en un diccionario serializable por xml"
        dic = {
            'Tipo_doc': self.tipo_doc, 'Nro_doc':  self.nro_doc,
            'Zona': self.zona,
            'Tipo_cbte': self.tipo_cbte, 'Fecha_cbte': self.fecha_cbte,
            'Punto_vta': self.punto_vta, 'Cbte_nro': self.cbte_nro,
            'Imp_total': self.imp_total, 'Imp_tot_conc': self.imp_tot_conc,
            'Imp_neto': self.imp_neto, 'Impto_liq': self.impto_liq,
            'Impto_liq_rni': self.impto_liq_rni, 'Imp_op_ex': self.imp_op_ex,
            'Imp_perc': self.imp_perc, 'Imp_iibb': self.imp_iibb,
            'Imp_perc_mun': self.imp_perc_mun, 'Imp_internos': self.imp_internos,
            'Imp_moneda_Id': self.imp_moneda_id, 'Imp_moneda_ctz': self.imp_moneda_ctz,
            'Items': [{'Item': item.to_dict()} for item in self.items],
            }
        if completo:
            dic.update({'nombre': self.nombre, 'categoria': self.categoria,
                        'domicilio': self.domicilio, 'localidad': self.localidad, 'provincia': self.provincia,
                        'email': self.email, 'telefono': self.telefono,
                        'observaciones': self.observaciones,
                        })
        return dic


class ItemBF:
    "Item de Factura Bono Fiscal"
    # valores por defecto:
    pro_codigo_ncm = ""
    pro_codigo_sec = ""
    pro_ds = ""
    pro_qty = 0.0
    pro_umed = 1
    pro_precio_uni = 0.0
    imp_bonif = 0.0
    imp_total = 0.0
    iva_id = 5

    def __init__(self, ncm="", sec="", ds="", qty=1, umed=7, precio=0.00, bonif=0.00, iva_id=5, imp_total=0.00, **kwargs):
        "Constructor" #TODO: revisar cálculo de iva y totales
        self.pro_codigo_ncm = ncm
        self.pro_codigo_sec = sec
        self.pro_ds = ds
        self.pro_qty = qty
        self.pro_umed = umed
        self.pro_precio_uni = precio
        self.imp_bonif = bonif
        self.imp_total = imp_total or (precio*qty - bonif)
        self.iva_id = iva_id

    def to_dict(self):
        "Convierte el objeto item en un diccionario serializable por xml"
        return {
            "Pro_codigo_ncm": self.pro_codigo_ncm,
            "Pro_codigo_sec": self.pro_codigo_sec,
            "Pro_ds": self.pro_ds,
            "Pro_qty": self.pro_qty,
            "Pro_umed": self.pro_umed,
            "Pro_precio_uni": self.pro_precio_uni,
            "Imp_bonif": self.imp_bonif,
            "Imp_total": self.imp_total,
            "Iva_id": self.iva_id,
            }

def main():
    "Función principal de pruebas (obtener CAE)"
    import os, time
    import sys, codecs, locale
    if sys.stdout.encoding is None:
        sys.stdout = codecs.getwriter("latin1")(sys.stdout,"replace");

    url = len(sys.argv)>2 and sys.argv[2].startswith("http") and sys.argv[2] or WSBFEURL

    # cliente soap del web service
    client = SoapClient(url,
        action = SOAP_ACTION,
        namespace = SOAP_NS,
        trace = "--trace" in sys.argv)

    # obteniendo el TA
    TA = "TA.xml"
    if not os.path.exists(TA) or os.path.getmtime(TA)+(60*60*5)<time.time():
        import wsaa
        tra = wsaa.create_tra(service="wsbfe")
        cms = wsaa.sign_tra(tra,"reingart.crt","reingart.key")
        ta_string = wsaa.call_wsaa(cms)
        open("TA.xml","w").write(ta_string)
    ta_string=open(TA).read()
    ta = SimpleXMLElement(ta_string)
    token = str(ta.credentials.token)
    sign = str(ta.credentials.sign)
    # fin TA

    CUIT = len(sys.argv)>1 and sys.argv[1] or "20267565393"

    if "--dummy" in sys.argv:
        print(dummy(client))

    # Recuperar parámetros:
    if "--params" in sys.argv:

        print("=== Monedas ===")
        monedas = get_param_mon(client, token, sign, CUIT)
        for m in monedas:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % m)
        monedas = dict([(m['id'],m['ds']) for m in monedas])

        print("=== NCM ===")
        productos = get_param_ncm(client, token, sign, CUIT)
        for p in productos:
            print("||%(codigo)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % p)
        productos = dict([(m['codigo'],m['ds']) for m in productos])

        print("=== Tipos de Comprobante ===")
        cbtes = get_param_tipo_cbte(client, token, sign, CUIT)
        for c in cbtes:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % c)
        cbtes = dict([(c['id'],c['ds']) for c in cbtes])

        print("=== Tipos de IVA ===")
        ivas = get_param_tipo_iva(client, token, sign, CUIT)
        for i in ivas:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % i)
        ivas = dict([(i['id'],i['ds']) for i in ivas])

        print("=== Unidades de Medida ===")
        umedidas = get_param_umed(client, token, sign, CUIT)
        for u in umedidas:
            print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % u)
        umeds = dict([(u['id'],u['ds']) for u in umedidas])

        ##zonas = get_param_zonas(client, token, sign, CUIT)
        # print(dict([(z['id'],z['ds']) for z in zonas]))

    else:

        monedas = {'DOL': u'D\xf3lar Estadounidense', 'PES': u'Pesos Argentinos', '010': u'Pesos Mejicanos', '011': u'Pesos Uruguayos', '012': u'Real', '014': u'Coronas Danesas', '015': u'Coronas Noruegas', '016': u'Coronas Suecas', '019': u'Yens', '018': u'D\xf3lar Canadiense', '033': u'Peso Chileno', '056': u'Forint (Hungr\xeda)', '031': u'Peso Boliviano', '036': u'Sucre Ecuatoriano', '051': u'D\xf3lar de Hong Kong', '034': u'Rand Sudafricano', '053': u'D\xf3lar de Jamaica', '057': u'Baht (Tailandia)', '043': u'Balboas Paname\xf1as', '042': u'Peso Dominicano', '052': u'D\xf3lar de Singapur', '032': u'Peso Colombiano', '035': u'Nuevo Sol Peruano', '061': u'Zloty Polaco', '060': u'Euro', '063': u'Lempira Hondure\xf1a', '062': u'Rupia Hind\xfa', '064': u'Yuan (Rep. Pop. China)', '025': u'Dinar Yugoslavo', '002': u'D\xf3lar Libre EEUU', '027': u'Dracma Griego', '026': u'D\xf3lar Australiano', '007': u'Florines Holandeses', '023': u'Bol\xedvar Venezolano', '047': u'Riyal Saudita', '046': u'Libra Egipcia', '045': u'Dirham Marroqu\xed', '044': u'C\xf3rdoba Nicarag\xfcense', '029': u'G\xfcaran\xed', '028': u'Flor\xedn (Antillas Holandesas)', '054': u'D\xf3lar de Taiwan', '040': u'Lei Rumano', '024': u'Corona Checa', '030': u'Shekel (Israel)', '021': u'Libra Esterlina', '055': u'Quetzal Guatemalteco', '059': u'Dinar Kuwaiti'}
        comprobantes = {1: u'Factura A', 2: u'Nota de D\xe9bito A', 3: u'Nota de Cr\xe9dito A', 4: u'Recibo A', 6: u'Factura B', 7: u'Nota de D\xe9bito B', 8: u'Nota de Cr\xe9dito B', 9: u'Recibo B', 11: u'Factura C', 12: u'Nota de D\xe9bito C', 13: u'Nota de Cr\xe9dito C', 15: u'Recibo C', 51: u'Factura M', 52: u'Nota de D\xe9bito M', 53: u'Nota de Cr\xe9dito M', 54: u'Recibo M'}
        ivas = {1: u'No gravado', 2: u'Exento', 3: u'0%', 4: u'10.5%', 5: u'21%', 6: u'27%'}
        umeds = {1: u'kilogramos', 2: u'metros', 3: u'metros cuadrados', 4: u'metros c\xfabicos', 5: u'litros', 6: u'1000 kWh', 7: u'unidades', 8: u'pares', 9: u'docenas', 10: u'quilates', 11: u'millares', 14: u'gramos', 15: u'milimetros', 16: u'mm c\xfabicos', 17: u'kil\xf3metros', 18: u'hectolitros', 20: u'cent\xedmetros', 25: u'jgo. pqt. mazo naipes', 27: u'cm c\xfabicos', 29: u'toneladas', 30: u'dam c\xfabicos', 31: u'hm c\xfabicos', 32: u'km c\xfabicos', 33: u'microgramos', 34: u'nanogramos', 35: u'picogramos', 41: u'miligramos', 47: u'mililitros', 48: u'curie', 49: u'milicurie', 50: u'microcurie', 51: u'uiacthor', 52: u'muiacthor', 53: u'kg base', 54: u'gruesa', 61: u'kg bruto', 62: u'uiactant', 63: u'muiactant', 64: u'uiactig', 65: u'muiactig', 66: u'kg activo', 67: u'gramo activo', 68: u'gramo base', 96: u'packs', 97: u'hormas', 98: u'bonificaci\xf2n', 99: u'otras unidades'}

    # ncm=7308.10.00, 7308.20.00

    if "--prueba" in sys.argv:

        # recupero ultimo comprobante y id
        tipo_cbte = 1
        punto_vta = 5
        cbte_nro, cbte_fecha, events = get_last_cmp(client, token, sign, CUIT, tipo_cbte, punto_vta)
        id, events = get_last_id(client, token, sign, CUIT)

        #cbte_nro = 1
        #id = 1002 # 99000000000098

        # creo una factura de prueba
        f = FacturaBF()
        f.punto_vta = punto_vta
        f.cbte_nro = cbte_nro+1
        f.imp_moneda_id = 'PES'
        f.fecha_cbte = date('Ymd')
        it = ItemBF(ncm='7308.10.00', sec='', ds=u'prueba Anafe económico', qty=2.0, precio=100.0, bonif=0.0, iva_id=5)
        f.add_item(it)
        it = ItemBF(ncm='7308.20.00', sec='', ds='prueba 2', qty=4.0, precio=50.0, bonif=10.0, iva_id=5)
        f.add_item(it)
        print(f.to_dict())

        try:
            # autorizar...
            auth, events = authorize(client, token, sign, CUIT, id=id+1, factura=f.to_dict())
            cae = auth['cae']
            print("auth", auth)
            print("events", events)
            auth, events = get_cmp(client, token, sign, CUIT, f.tipo_cbte, f.punto_vta, f.cbte_nro)
            print("get_cmp: auth", auth)
            print("get_cmp: events", events)
        except:
            raise
            # l= client.xml_request.splitlines()
            # print(l[4][12.0.1.00])
            # import pdb; pdb.set_trace()

    sys.exit(0)

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener CAE, código de autorización de impresión o electrónico,
del web service WSFEXv1 de AFIP (Factura Electrónica Exportación Versión 1)
según RG2758/2010 (Registros Especiales Aduaneros) y RG3689/14 (servicios)
http://www.sistemasagiles.com.ar/trac/wiki/FacturaElectronicaExportacion
"""

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2011-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.08e"

import datetime
import sys
from .utils import inicializar_y_capturar_excepciones, BaseWS, get_install_dir
import logging

log = logging.getLogger()

HOMO = False
WSDL = "https://wswhomo.afip.gov.ar/wsfexv1/service.asmx?WSDL"


class WSFEXv1(BaseWS):
    "Interfaz para el WebService de Factura Electrónica Exportación Versión 1"
    _public_methods_ = [
        'CrearFactura', 'AgregarItem', 'Authorize', 'GetCMP',
        'AgregarPermiso', 'AgregarCmpAsoc', 'GetParamMon',
        'GetParamTipoCbte', 'GetParamTipoExpo', 'GetParamIdiomas',
        'GetParamUMed', 'GetParamIncoterms', 'GetParamDstPais',
        'GetParamDstCUIT', 'GetParamIdiomas', 'GetParamIncoterms',
        'GetParamDstCUIT', 'GetParamPtosVenta', 'GetParamCtz', 'LoadTestXML',
        'AnalizarXml', 'ObtenerTagXml', 'DebugLog', 'SetParametros',
        'SetTicketAcceso', 'GetParametro', 'GetLastCMP', 'GetLastID',
        'Dummy', 'Conectar', 'SetTicketAcceso'
    ]
    _public_attrs_ = [
        'Token', 'Sign', 'Cuit', 'AppServerStatus', 'DbServerStatus',
        'AuthServerStatus', 'XmlRequest', 'XmlResponse', 'Version',
        'Resultado', 'Obs', 'Reproceso', 'CAE', 'Vencimiento', 'Eventos',
        'ErrCode', 'ErrMsg', 'FchVencCAE', 'Excepcion', 'LanzarExcepciones',
        'Traceback', "InstallDir", 'PuntoVenta', 'CbteNro', 'FechaCbte',
        'ImpTotal'
    ]

    _reg_progid_ = "WSFEXv1"
    _reg_clsid_ = "{8106F039-D132-4F87-8AFE-ADE47B5503D4}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
    factura = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.CbteNro = self.FechaCbte = self.PuntoVenta = self.ImpTotal = None
        self.InstallDir = INSTALL_DIR
        self.FchVencCAE = ""              # retrocompatibilidad

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'FEXErr' in ret:
            errores = [ret['FEXErr']]
            for error in errores:
                self.Errores.append("%s: %s" % (
                    error['ErrCode'],
                    error['ErrMsg'],
                    ))
            self.ErrCode = ' '.join(
                [str(error['ErrCode']) for error in errores])
            self.ErrMsg = '\n'.join(self.Errores)
        if 'FEXEvents' in ret:
            events = [ret['FEXEvents']]
            self.Eventos = [
                '%s: %s' % (evt['EventCode'],
                            evt.get('EventMsg', "")) for evt in events]

    def CrearFactura(
        self, tipo_cbte=19, punto_vta=1, cbte_nro=0, fecha_cbte=None,
        imp_total=0.0, tipo_expo=1, permiso_existente="N", pais_dst_cmp=None,
        nombre_cliente="", cuit_pais_cliente="", domicilio_cliente="",
        id_impositivo="", moneda_id="PES", moneda_ctz=1.0,
        obs_comerciales="", obs_generales="", forma_pago="", incoterms="",
        idioma_cbte=7, incoterms_ds=None, **kwargs
    ):
        "Creo un objeto factura (interna)"
        # Creo una factura electronica de exportación

        fact = {
            'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
            'cbte_nro': cbte_nro, 'fecha_cbte': fecha_cbte,
            'tipo_doc': 80, 'nro_doc':  cuit_pais_cliente,
            'imp_total': imp_total,
            'permiso_existente': permiso_existente,
            'pais_dst_cmp': pais_dst_cmp,
            'nombre_cliente': nombre_cliente,
            'domicilio_cliente': domicilio_cliente,
            'id_impositivo': id_impositivo,
            'moneda_id': moneda_id, 'moneda_ctz': moneda_ctz,
            'obs_comerciales': obs_comerciales,
            'obs_generales': obs_generales,
            'forma_pago': forma_pago,
            'incoterms': incoterms,
            'incoterms_ds': incoterms_ds,
            'tipo_expo': tipo_expo,
            'idioma_cbte': idioma_cbte,
            'cbtes_asoc': [],
            'permisos': [],
            'detalles': [],
        }
        self.factura = fact

        return True

    def AgregarItem(
        self, codigo, ds, qty, umed, precio, importe, bonif=None, **kwargs
    ):
        "Agrego un item a una factura (interna)"
        # Nota: no se calcula total (debe venir calculado!)
        self.factura['detalles'].append({
                'codigo': codigo,
                'ds': ds,
                'qty': qty,
                'umed': umed,
                'precio': precio,
                'bonif': bonif,
                'importe': importe,
                })
        return True

    def AgregarPermiso(self, id_permiso, dst_merc, **kwargs):
        "Agrego un permiso a una factura (interna)"
        self.factura['permisos'].append({
                'id_permiso': id_permiso,
                'dst_merc': dst_merc,
                })
        return True

    def AgregarCmpAsoc(
        self,
        cbte_tipo=19,
        cbte_punto_vta=0,
        cbte_nro=0,
        cbte_cuit=None,
        **kwargs
    ):
        "Agrego un comprobante asociado a una factura (interna)"
        self.factura['cbtes_asoc'].append({
            'cbte_tipo': cbte_tipo, 'cbte_punto_vta': cbte_punto_vta,
            'cbte_nro': cbte_nro, 'cbte_cuit': cbte_cuit})
        return True

    @inicializar_y_capturar_excepciones
    def Authorize(self, id):
        "Autoriza la factura cargada en memoria"
        f = self.factura
        ret = self.client.FEXAuthorize(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            Cmp={
                'Id': id,
                'Fecha_cbte': f['fecha_cbte'],
                'Cbte_Tipo': f['tipo_cbte'],
                'Punto_vta': f['punto_vta'],
                'Cbte_nro': f['cbte_nro'],
                'Tipo_expo': f['tipo_expo'],
                'Permiso_existente': f['permiso_existente'],
                'Permisos': f['permisos'] and [
                    {'Permiso': {
                        'Id_permiso': p['id_permiso'],
                        'Dst_merc': p['dst_merc'],
                    }} for p in f['permisos']] or None,
                'Dst_cmp': f['pais_dst_cmp'],
                'Cliente': f['nombre_cliente'],
                'Cuit_pais_cliente': f['nro_doc'],
                'Domicilio_cliente': f['domicilio_cliente'],
                'Id_impositivo': f['id_impositivo'],
                'Moneda_Id': f['moneda_id'],
                'Moneda_ctz': f['moneda_ctz'],
                'Obs_comerciales': f['obs_comerciales'],
                'Imp_total': f['imp_total'],
                'Obs': f['obs_generales'],
                'Cmps_asoc': f['cbtes_asoc'] and [
                    {'Cmp_asoc': {
                        'Cbte_tipo': c['cbte_tipo'],
                        'Cbte_punto_vta': c['cbte_punto_vta'],
                        'Cbte_nro': c['cbte_nro'],
                        'Cbte_cuit': c['cbte_cuit'],
                    }} for c in f['cbtes_asoc']] or None,
                'Forma_pago': f['forma_pago'],
                'Incoterms': f['incoterms'],
                'Incoterms_Ds': f['incoterms_ds'],
                'Idioma_cbte':  f['idioma_cbte'],
                'Items': [
                    {'Item': {
                        'Pro_codigo': d['codigo'],
                        'Pro_ds': d['ds'],
                        'Pro_qty': d['qty'],
                        'Pro_umed': d['umed'],
                        'Pro_precio_uni': d['precio'],
                        'Pro_bonificacion': d['bonif'],
                        'Pro_total_item': d['importe'],
                     }} for d in f['detalles']],
            })

        result = ret['FEXAuthorizeResult']
        self.__analizar_errores(result)
        if 'FEXResultAuth' in result:
            auth = result['FEXResultAuth']
            # Resultado: A: Aceptado, R: Rechazado
            self.Resultado = auth.get('Resultado', "")
            # Obs:
            self.Obs = auth.get('Motivos_Obs', "")
            self.Reproceso = auth.get('Reproceso', "")
            self.CAE = auth.get('Cae', "")
            self.CbteNro = auth.get('Cbte_nro', "")
            vto = str(auth.get('Fch_venc_Cae', ""))
            self.FchVencCAE = vto
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            return self.CAE

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        result = self.client.FEXDummy()['FEXDummyResult']
        self.__analizar_errores(result)
        self.AppServerStatus = str(result['AppServer'])
        self.DbServerStatus = str(result['DbServer'])
        self.AuthServerStatus = str(result['AuthServer'])
        return True

    @inicializar_y_capturar_excepciones
    def GetCMP(self, tipo_cbte, punto_vta, cbte_nro):
        "Recuperar los datos completos de un comprobante ya autorizado"
        ret = self.client.FEXGetCMP(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            Cmp={
                'Cbte_tipo': tipo_cbte,
                'Punto_vta': punto_vta,
                'Cbte_nro': cbte_nro,
            })
        result = ret['FEXGetCMPResult']
        self.__analizar_errores(result)
        if 'FEXResultGet' in result:
            resultget = result['FEXResultGet']
            # Obs, cae y fecha cae
            self.Obs = resultget.get('Obs') and \
                resultget['Obs'].strip(" ") or ''
            self.CAE = resultget.get('Cae', '')
            vto = str(resultget.get('Fch_venc_Cae', ''))
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            # .strftime("%Y/%m/%d")
            self.FechaCbte = resultget.get('Fecha_cbte', '')
            self.PuntoVenta = resultget['Punto_vta']  # 4000
            self.Resultado = resultget.get('Resultado', '')
            self.CbteNro = resultget['Cbte_nro']
            self.ImpTotal = str(resultget['Imp_total'])
            return self.CAE
        else:
            return 0

    @inicializar_y_capturar_excepciones
    def GetLastCMP(self, tipo_cbte, punto_vta):
        "Recuperar último número de comprobante emitido"
        ret = self.client.FEXGetLast_CMP(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
                'Cbte_Tipo': tipo_cbte,
                'Pto_venta': punto_vta,
            }
        )
        result = ret['FEXGetLast_CMPResult']
        self.__analizar_errores(result)
        if 'FEXResult_LastCMP' in result:
            resultget = result['FEXResult_LastCMP']
            self.CbteNro = resultget.get('Cbte_nro')
            # .strftime("%Y/%m/%d")
            self.FechaCbte = resultget.get('Cbte_fecha')
            return self.CbteNro

    @inicializar_y_capturar_excepciones
    def GetLastID(self):
        "Recuperar último número de transacción (ID)"
        ret = self.client.FEXGetLast_ID(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetLast_IDResult']
        self.__analizar_errores(result)
        if 'FEXResultGet' in result:
            resultget = result['FEXResultGet']
            return resultget.get('Id')

    @inicializar_y_capturar_excepciones
    def GetParamUMed(self, sep="|"):
        ret = self.client.FEXGetPARAM_UMed(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_UMedResult']
        self.__analizar_errores(result)

        umeds = []  # unidades de medida
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_UMed']
            try:
                umed = {
                    'id': u.get('Umed_Id'),
                    'ds': u.get('Umed_Ds'),
                    'vig_desde': u.get('Umed_vig_desde'),
                    'vig_hasta': u.get('Umed_vig_hasta')
                }
            except Exception as e:
                print(e)
                if u is None:
                    # <ClsFEXResponse_UMed xsi:nil="true"/> WTF!
                    umed = {
                        'id': '',
                        'ds': '',
                        'vig_desde': '',
                        'vig_hasta': ''
                    }
                    # import pdb; pdb.set_trace()
                    # print(u)

            umeds.append(umed)
        if sep:
            return [
                ("\t%(id)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in umeds
            ]
        else:
            return umeds

    @inicializar_y_capturar_excepciones
    def GetParamMon(self, sep="|"):
        ret = self.client.FEXGetPARAM_MON(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_MONResult']
        self.__analizar_errores(result)

        mons = []  # monedas
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_Mon']
            try:
                mon = {
                    'id': u.get('Mon_Id'),
                    'ds': u.get('Mon_Ds'),
                    'vig_desde': u.get('Mon_vig_desde'),
                    'vig_hasta': u.get('Mon_vig_hasta')
                }
            except Exception as e:
                print(e)
                if u is None:
                    # <ClsFEXResponse_UMed xsi:nil="true"/> WTF!
                    mon = {
                        'id': '',
                        'ds': '',
                        'vig_desde': '',
                        'vig_hasta': ''
                    }
                    # import pdb; pdb.set_trace()
                    # print(u)

            mons.append(mon)
        if sep:
            return [
                ("\t%(id)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in mons
            ]
        else:
            return mons

    @inicializar_y_capturar_excepciones
    def GetParamDstPais(self, sep="|"):
        "Recuperador de valores referenciales de códigos de Países"
        ret = self.client.FEXGetPARAM_DST_pais(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_DST_paisResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_DST_pais']
            try:
                r = {
                    'codigo': u.get('DST_Codigo'),
                    'ds': u.get('DST_Ds'),
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def GetParamDstCUIT(self, sep="|"):
        "Recuperar lista de valores referenciales de CUIT de Países"
        ret = self.client.FEXGetPARAM_DST_CUIT(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_DST_CUITResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_DST_cuit']
            try:
                r = {
                    'codigo': u.get('DST_CUIT'),
                    'ds': u.get('DST_Ds'),
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def GetParamTipoCbte(self, sep="|"):
        """Recuperador de valores referenciales de códigos de Tipo de
        comprobantes"""
        ret = self.client.FEXGetPARAM_Cbte_Tipo(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_Cbte_TipoResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_Cbte_Tipo']
            try:
                r = {
                    'codigo': u.get('Cbte_Id'),
                    'ds': u.get('Cbte_Ds').replace('\n', '').replace('\r', ''),
                    'vig_desde': u.get('Cbte_vig_desde'),
                    'vig_hasta': u.get('Cbte_vig_hasta')
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def GetParamTipoExpo(self, sep="|"):
        """Recuperador de valores referenciales de códigos de Tipo de
        exportación"""
        ret = self.client.FEXGetPARAM_Tipo_Expo(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_Tipo_ExpoResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_Tex']
            try:
                r = {
                    'codigo': u.get('Tex_Id'),
                    'ds': u.get('Tex_Ds'),
                    'vig_desde': u.get('Tex_vig_desde'),
                    'vig_hasta': u.get('Tex_vig_hasta')
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def GetParamIdiomas(self, sep="|"):
        "Recuperar lista de valores referenciales de códigos de Idiomas"
        ret = self.client.FEXGetPARAM_Idiomas(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_IdiomasResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_Idi']
            try:
                r = {
                    'codigo': u.get('Idi_Id'),
                    'ds': u.get('Idi_Ds'),
                    'vig_desde': u.get('Idi_vig_hasta'),
                    'vig_hasta': u.get('Idi_vig_desde')
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    def GetParamIncoterms(self, sep="|"):
        "Recuperar lista de valores referenciales de Incoterms"
        ret = self.client.FEXGetPARAM_Incoterms(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit,
            }
        )
        result = ret['FEXGetPARAM_IncotermsResult']
        self.__analizar_errores(result)

        ret = []
        for u in result['FEXResultGet']:
            u = u['ClsFEXResponse_Inc']
            try:
                r = {
                    'codigo': u.get('Inc_Id'),
                    'ds': u.get('Inc_Ds'),
                    'vig_desde': u.get('Inc_vig_hasta'),
                    'vig_hasta': u.get('Inc_vig_desde')
                }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [
                ("\t%(codigo)s\t%(ds)s\t%(vig_desde)s\t%(vig_hasta)s\t" % it)
                .replace("\t", sep) for it in ret
            ]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def GetParamCtz(self, moneda_id):
        "Recuperador de cotización de moneda"
        ret = self.client.FEXGetPARAM_Ctz(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            Mon_id=moneda_id,
            )
        self.__analizar_errores(ret['FEXGetPARAM_CtzResult'])
        res = ret['FEXGetPARAM_CtzResult'].get('FEXResultGet')
        if res:
            ctz = str(res.get('Mon_ctz', ""))
        else:
            ctz = ''
        return ctz

    @inicializar_y_capturar_excepciones
    def GetParamPtosVenta(self, sep="|"):
        "Recupera el listado de los puntos de venta para exportacion y estado"
        ret = self.client.FEXGetPARAM_PtoVenta(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        self.__analizar_errores(ret['FEXGetPARAM_PtoVentaResult'])
        res = ret['FEXGetPARAM_PtoVentaResult'].get('FEXResultGet')
        log.warn("FEXGetPARAM_PtoVentaResult Result: " + str(res))
        ret = []
        for pu in res:
            log.warn("pu Result: " + str(pu))
            p = pu['ClsFEXResponse_PtoVenta']
            log.warn("p Result: " + str(p))
            try:
                r = {
                    'nro': p.get('Pve_Nro'),
                    'baja': p.get('Pve_FchBaj'),
                    'bloqueado': p.get('Pve_Bloqueado'),
                }
            except Exception as e:
                print(e)
            ret.append(r)
        log.warn("FEXGetPARAM_PtoVentaResult ret: " + str(ret))
        return [
            ("%(nro)s\tBloqueado:%(bloqueado)s\tFchBaja:%(baja)s" % rt)
            .replace("\t", sep) for rt in ret
        ]


class WSFEX(WSFEXv1):
    "Wrapper para retrocompatibilidad con WSFEX"

    _reg_progid_ = "WSFEX"
    _reg_clsid_ = "{B3C8D3D3-D5DA-44C9-B003-11845803B2BD}"

    def __init__(self):
        WSFEXv1.__init__(self)
        self.Version = "%s %s WSFEXv1" % (
            __version__, HOMO and 'Homologación' or '')

    def Conectar(self, url="", proxy=""):
        # Ajustar URL de V0 a V1:
        if url in ("https://wswhomo.afip.gov.ar/wsfex/service.asmx",
                   "http://wswhomo.afip.gov.ar/WSFEX/service.asmx"):
            url = "https://wswhomo.afip.gov.ar/wsfexv1/service.asmx"
        elif url in ("https://servicios1.afip.gov.ar/wsfex/service.asmx",
                     "http://servicios1.afip.gov.ar/WSFEX/service.asmx"):
            url = "https://servicios1.afip.gov.ar/wsfexv1/service.asmx"
        return WSFEXv1.Conectar(self, cache=None, wsdl=url, proxy=proxy)


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSFEXv1.InstallDir = get_install_dir()


def p_assert_eq(a,b):
    print(a, a==b and '==' or '!=', b)


if __name__ == "__main__":

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSFEXv1)
        if '--wsfex' in sys.argv:
            win32com.server.register.UseCommandLine(WSFEX)
    #elif "/Automate" in sys.argv:
    #    # MS seems to like /automate to run the class factories.
    #    import win32com.server.localserver
    #    #win32com.server.localserver.main()
    #    # start the server.
    #    win32com.server.localserver.serve([WSFEXv1._reg_clsid_])
    else:

        # Crear objeto interface Web Service de Factura Electrónica de Exportación
        wsfexv1 = WSFEXv1()
        # Setear token y sing de autorización (pasos previos)

        # obteniendo el TA para pruebas
        from wsaa import WSAA
        ta = WSAA().Autenticar("wsfex", "reingart.crt", "reingart.key")
        wsfexv1.SetTicketAcceso(ta)

        # CUIT del emisor (debe estar registrado en la AFIP)
        wsfexv1.Cuit = "20267565393"

        # Conectar al Servicio Web de Facturación (producción u homologación)
        if "--prod" in sys.argv:
            wsdl = "https://servicios1.afip.gov.ar/wsfexv1/service.asmx?WSDL"
        else:
            wsdl = "https://wswhomo.afip.gov.ar/wsfexv1/service.asmx?WSDL"
        cache = proxy = ""
        wrapper = "httplib2"
        cacert = open("conf/afip_ca_info.crt").read()
        ok = wsfexv1.Conectar(cache, wsdl, proxy, wrapper, cacert)

        if '--dummy' in sys.argv:
            #wsfexv1.LanzarExcepciones = False
            print(wsfexv1.Dummy())
            print(wsfexv1.XmlRequest)
            print(wsfexv1.XmlResponse)

        if "--prueba" in sys.argv:
            try:
                # Establezco los valores de la factura a autorizar:
                tipo_cbte = '--nc' in sys.argv and 21 or 19 # FC/NC Expo (ver tabla de parámetros)
                punto_vta = 7
                # Obtengo el último número de comprobante y le agrego 1
                cbte_nro = int(wsfexv1.GetLastCMP(tipo_cbte, punto_vta)) + 1
                fecha_cbte = datetime.datetime.now().strftime("%Y%m%d")
                tipo_expo = 1 # tipo de exportación (ver tabla de parámetros)
                permiso_existente = (tipo_cbte not in (20, 21) or tipo_expo!=1) and "S" or ""
                print("permiso_existente", permiso_existente)
                dst_cmp = 203 # país destino
                cliente = "Joao Da Silva"
                cuit_pais_cliente = "50000000016"
                domicilio_cliente = u"Rúa Ñ°76 km 34.5 Alagoas"
                id_impositivo = "PJ54482221-l"
                moneda_id = "DOL" # para reales, "DOL" o "PES" (ver tabla de parámetros)
                moneda_ctz = "8.00" # wsfexv1.GetParamCtz('DOL') <- no funciona
                obs_comerciales = "Observaciones comerciales"
                obs = "Sin observaciones"
                forma_pago = "30 dias"
                incoterms = "FOB" # (ver tabla de parámetros)
                incoterms_ds = "Flete a Bordo"
                idioma_cbte = 1 # (ver tabla de parámetros)
                imp_total = "250.00"

                # Creo una factura (internamente, no se llama al WebService):
                ok = wsfexv1.CrearFactura(tipo_cbte, punto_vta, cbte_nro, fecha_cbte,
                        imp_total, tipo_expo, permiso_existente, dst_cmp,
                        cliente, cuit_pais_cliente, domicilio_cliente,
                        id_impositivo, moneda_id, moneda_ctz,
                        obs_comerciales, obs, forma_pago, incoterms,
                        idioma_cbte, incoterms_ds)

                # Agrego un item:
                codigo = "PRO1"
                ds = "Producto Tipo 1 Exportacion MERCOSUR ISO 9001"
                qty = 2
                precio = "150.00"
                umed = 1 # Ver tabla de parámetros (unidades de medida)
                bonif = "50.00"
                imp_total = "250.00" # importe total final del artículo
                # lo agrego a la factura (internamente, no se llama al WebService):
                ok = wsfexv1.AgregarItem(codigo, ds, qty, umed, precio, imp_total, bonif)
                ok = wsfexv1.AgregarItem(codigo, ds, qty, umed, precio, imp_total, bonif)
                ok = wsfexv1.AgregarItem(codigo, ds, 0, 99, 0, -float(imp_total), 0)

                # Agrego un permiso (ver manual para el desarrollador)
                if permiso_existente:
                    id = "99999AAXX999999A"
                    dst = 225 # país destino de la mercaderia
                    ok = wsfexv1.AgregarPermiso(id, dst)

                # Agrego un comprobante asociado (solo para N/C o N/D)
                if tipo_cbte in (20,21):
                    cbteasoc_tipo = 19
                    cbteasoc_pto_vta = 2
                    cbteasoc_nro = 1234
                    cbteasoc_cuit = 20111111111
                    wsfexv1.AgregarCmpAsoc(cbteasoc_tipo, cbteasoc_pto_vta, cbteasoc_nro, cbteasoc_cuit)

                ##id = "99000000000100" # número propio de transacción
                # obtengo el último ID y le adiciono 1
                # (advertencia: evitar overflow y almacenar!)
                id = long(wsfexv1.GetLastID()) + 1

                # Llamo al WebService de Autorización para obtener el CAE
                cae = wsfexv1.Authorize(id)

                print("Comprobante", tipo_cbte, wsfexv1.CbteNro)
                print("Resultado", wsfexv1.Resultado)
                print("CAE", wsfexv1.CAE)
                print("Vencimiento", wsfexv1.Vencimiento)

                if wsfexv1.Resultado and False:
                    print(wsfexv1.client.help("FEXGetCMP").encode("latin1"))
                    wsfexv1.GetCMP(tipo_cbte, punto_vta, cbte_nro)
                    print("CAE consulta", wsfexv1.CAE, wsfexv1.CAE==cae)
                    print("NRO consulta", wsfexv1.CbteNro, wsfexv1.CbteNro==cbte_nro)
                    print("TOTAL consulta", wsfexv1.ImpTotal, wsfexv1.ImpTotal==imp_total)

            except Exception as e:
                print(wsfexv1.XmlRequest)
                print(wsfexv1.XmlResponse)
                print(wsfexv1.ErrCode)
                print(wsfexv1.ErrMsg)
                print(wsfexv1.Excepcion)
                print(wsfexv1.Traceback)
                raise

        if "--get" in sys.argv:
            wsfexv1.client.help("FEXGetCMP")
            tipo_cbte = 19
            punto_vta = 7
            cbte_nro = wsfexv1.GetLastCMP(tipo_cbte, punto_vta)

            wsfexv1.GetCMP(tipo_cbte, punto_vta, cbte_nro)

            print("FechaCbte = ", wsfexv1.FechaCbte)
            print("CbteNro = ", wsfexv1.CbteNro)
            print("PuntoVenta = ", wsfexv1.PuntoVenta)
            print("ImpTotal =", wsfexv1.ImpTotal)
            print("CAE = ", wsfexv1.CAE)
            print("Vencimiento = ", wsfexv1.Vencimiento)

            wsfexv1.AnalizarXml("XmlResponse")
            p_assert_eq(wsfexv1.ObtenerTagXml('Cae'), str(wsfexv1.CAE))
            p_assert_eq(wsfexv1.ObtenerTagXml('Fecha_cbte'), wsfexv1.FechaCbte)
            p_assert_eq(wsfexv1.ObtenerTagXml('Moneda_Id'), "DOL")
            p_assert_eq(wsfexv1.ObtenerTagXml('Moneda_ctz'), "8")
            p_assert_eq(wsfexv1.ObtenerTagXml('Id_impositivo'), "PJ54482221-l")

        if "--params" in sys.argv:
            import codecs, locale
            sys.stdout = codecs.getwriter('latin1')(sys.stdout);

            print("=== Incoterms ===")
            idiomas = wsfexv1.GetParamIncoterms(sep="||")
            for idioma in idiomas:
                print(idioma)

            print("=== Idiomas ===")
            idiomas = wsfexv1.GetParamIdiomas(sep="||")
            for idioma in idiomas:
                print(idioma)

            print("=== Tipos Comprobantes ===")
            tipos = wsfexv1.GetParamTipoCbte(sep=False)
            for t in tipos:
                print("||%(codigo)s||%(ds)s||" % t)

            print("=== Tipos Expo ===")
            tipos = wsfexv1.GetParamTipoExpo(sep=False)
            for t in tipos:
                print("||%(codigo)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % t)
            #umeds = dict([(u.get('id', ""),u.get('ds', "")) for u in umedidas])

            print("=== Monedas ===")
            mons = wsfexv1.GetParamMon(sep=False)
            for m in mons:
                print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % m)
            #umeds = dict([(u.get('id', ""),u.get('ds', "")) for u in umedidas])

            print("=== Unidades de medida ===")
            umedidas = wsfexv1.GetParamUMed(sep=False)
            for u in umedidas:
                print("||%(id)s||%(ds)s||%(vig_desde)s||%(vig_hasta)s||" % u)
            umeds = dict([(u.get('id', ""),u.get('ds', "")) for u in umedidas])

            print(u"=== Código Pais Destino ===")
            ret = wsfexv1.GetParamDstPais(sep=False)
            for r in ret:
                print("||%(codigo)s||%(ds)s||" % r)

            print(u"=== CUIT Pais Destino ===")
            ret = wsfexv1.GetParamDstCUIT(sep=False)
            for r in ret:
                print("||%(codigo)s||%(ds)s||" % r)

        if "--ctz" in sys.argv:
            print(wsfexv1.GetParamCtz('DOL'))

        if "--ptosventa" in sys.argv:
            print(wsfexv1.GetParamPtosVenta())




#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Herramienta para procesar y consultar el Padrón Unico de Contribuyentes AFIP"

# Documentación e información adicional:
#    http://www.sistemasagiles.com.ar/trac/wiki/PadronContribuyentesAFIP

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2014 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.05b"


import json
import os
import shelve
import sqlite3
import urllib.request, urllib.error, urllib.parse
import zipfile
from email.utils import formatdate
import sys
from .utils import (
    leer,
    escribir,
    N,
    A,
    I,
    get_install_dir,
    safe_console,
    inicializar_y_capturar_excepciones_simple,
    WebClient
)

import logging
log = logging.getLogger()


# formato y ubicación archivo completo de la condición tributaria según RG 1817

FORMATO = [
    ("nro_doc", 11, N, ""),
    ("denominacion", 30, A, ""),
    ("imp_ganancias", 2, A,	"'NI', 'AC','EX', 'NC'"),
    ("imp_iva", 2, A, "'NI' , 'AC','EX','NA','XN','AN'"),
    ("monotributo", 2, A, "'NI', 'Codigo categoria tributaria'"),
    ("integrante_soc", 1, A, "'N' , 'S'"),
    ("empleador", 1, A, "'N', 'S'"),
    ("actividad_monotributo", 2, A, ""),
    ("tipo_doc", 2, N, "80: CUIT, 96: DNI, etc."),
    ("cat_iva", 2, N, "1: RI, 4: EX, 5: CF, 6: MT, etc"),
    ("email", 250, A, ""),
    ]

# Mapeos constantes:

PROVINCIAS = {
    0: 'CIUDAD AUTONOMA BUENOS AIRES', 1: 'BUENOS AIRES',
    2: 'CATAMARCA', 3: 'CORDOBA', 4: 'CORRIENTES', 5: 'ENTRE RIOS', 6: 'JUJUY',
    7: 'MENDOZA', 8: 'LA RIOJA', 9: 'SALTA', 10: 'SAN JUAN', 11: 'SAN LUIS',
    12: 'SANTA FE', 13: 'SANTIAGO DEL ESTERO', 14: 'TUCUMAN', 16: 'CHACO',
    17: 'CHUBUT', 18: 'FORMOSA', 19: 'MISIONES', 20: 'NEUQUEN', 21: 'LA PAMPA',
    22: 'RIO NEGRO', 23: 'SANTA CRUZ', 24: 'TIERRA DEL FUEGO'
}

TIPO_CLAVE = {'CUIT': 80, 'CUIL': 86, 'CDI': 86, 'DNI': 96, 'Otro': 99}

DEBUG = True

URL = "http://www.afip.gob.ar/genericos/cInscripcion/archivos/"
"apellidoNombreDenominacion.zip"
URL_API = "https://soa.afip.gob.ar/"


class PadronAFIP():
    "Interfaz para consultar situación tributaria (Constancia de Inscripcion)"

    _public_methods_ = ['Buscar', 'Descargar', 'Procesar', 'Guardar',
                        'ConsultarDomicilios', 'Consultar', 'Conectar',
                        'DescargarConstancia', 'MostrarPDF',
                        'ObtenerTablaParametros',
                        ]
    _public_attrs_ = ['InstallDir', 'Traceback', 'Excepcion', 'Version',
                      'cuit', 'dni', 'denominacion', 'imp_ganancias',
                      'imp_iva', 'monotributo', 'integrante_soc', 'empleador',
                      'actividad_monotributo', 'cat_iva', 'domicilios',
                      'tipo_doc', 'nro_doc', 'LanzarExcepciones',
                      'tipo_persona', 'estado', 'impuestos', 'actividades',
                      'direccion', 'localidad', 'provincia', 'cod_postal',
                      'data', 'response']
    _readonly_attrs_ = _public_attrs_[3:-1]
    _reg_progid_ = "PadronAFIP"
    _reg_clsid_ = "{6206DF5E-3EEF-47E9-A532-CD81EBBAF3AA}"

    def __init__(self):
        self.db_path = os.path.join(self.InstallDir, "padron.db")
        self.Version = __version__
        # Abrir la base de datos
        self.db = sqlite3.connect(self.db_path)
        self.db.row_factory = sqlite3.Row
        self.cursor = self.db.cursor()
        self.LanzarExcepciones = False
        self.inicializar()
        self.client = None

    def inicializar(self):
        self.Excepcion = self.Traceback = ""
        self.cuit = self.dni = 0
        self.tipo_persona = ""                      # FISICA o JURIDICA
        self.tipo_doc = 0
        self.estado = ""                            # ACTIVO
        self.denominacion = ""
        self.direccion = self.localidad = self.provincia = self.cod_postal = ""
        self.domicilios = []
        self.impuestos = []
        self.actividades = []
        self.imp_iva = self.empleador = self.integrante_soc = self.cat_iva = ""
        self.monotributo = self.actividad_monotributo = ""
        self.data = {}
        self.response = ""

    @inicializar_y_capturar_excepciones_simple
    def Conectar(
        self, url=URL_API, proxy="", wrapper=None, cacert=None, trace=False
    ):
        self.client = WebClient(location=url, trace=trace, cacert=cacert)
        self.client.method = "GET"       # metodo RESTful predeterminado
        self.client.enctype = None       # no enviar body
        return True

    @inicializar_y_capturar_excepciones_simple
    def Descargar(self, url=URL, filename="padron.txt", proxy=None):
        "Descarga el archivo de AFIP, devuelve 200 o 304 si no fue modificado"
        proxies = {}
        if proxy:
            proxies['http'] = proxy
            proxies['https'] = proxy
            proxy_handler = urllib.request.ProxyHandler(proxies)
        log.warn(("Abriendo URL %s ..." % url))
        req = urllib.request.Request(url)
        if os.path.exists(filename):
            http_date = formatdate(timeval=os.path.getmtime(filename),
                                   localtime=False, usegmt=True)
            req.add_header('If-Modified-Since', http_date)
        try:
            web = urllib.request.urlopen(req)
        except urllib.error.HTTPError as e:
            if e.code == 304:
                log.warn(("No modificado desde", http_date))
                return 304
            else:
                raise
        # leer info del request:
        meta = web.info()
        lenght = float(meta['Content-Length'])
        date = meta['Last-Modified']
        tmp = open(filename + ".zip", "wb")
        log.warn("Guardando")
        size = 0
        p0 = None
        while True:
            p = int(size / lenght * 100)
            if p0 is None or p > p0:
                print(("Leyendo ... %0d %%" % p))
                p0 = p
            data = web.read(1024*100)
            size = size + len(data)
            if not data:
                print("Descarga Terminada!")
                break
            tmp.write(data)
        log.warn("Abriendo ZIP...")
        tmp.close()
        web.close()
        uf = open(filename + ".zip", "rb")
        zf = zipfile.ZipFile(uf)
        for fn in zf.namelist():
            log.warn(("descomprimiendo", fn))
            tf = open(filename, "wb")
            tf.write(zf.read(fn))
            tf.close()
        return 200

    @inicializar_y_capturar_excepciones_simple
    def Procesar(self, filename="padron.txt", borrar=False):
        "Analiza y crea la base de datos interna sqlite para consultas"
        f = open(filename, "r")
        keys = [k for k, l, t, d in FORMATO]
        # conversion a planilla csv (no usado)
        if False and not os.path.exists("padron.csv"):
            csvfile = open('padron.csv', 'wb')
            import csv
            wr = csv.writer(csvfile, delimiter=',',
                            quotechar='"', quoting=csv.QUOTE_MINIMAL)
            for i, l in enumerate(f):
                if i % 100000 == 0:
                    print(("Progreso: %d registros" % i))
                r = leer(l, FORMATO)
                row = [r[k] for k in keys]
                wr.writerow(row)
            csvfile.close()
        f.seek(0)
        if os.path.exists(self.db_path) and borrar:
            os.remove(self.db_path)
        if True:
            db = db = sqlite3.connect(self.db_path)
            c = db.cursor()
            c.execute("CREATE TABLE padron ("
                        "nro_doc INTEGER, "
                        "denominacion VARCHAR(30), "
                        "imp_ganancias VARCHAR(2), "
                        "imp_iva VARCHAR(2), "
                        "monotributo VARCHAR(1), "
                        "integrante_soc VARCHAR(1), "
                        "empleador VARCHAR(1), "
                        "actividad_monotributo VARCHAR(2), "
                        "tipo_doc INTEGER, "
                        "cat_iva INTEGER DEFAULT NULL, "
                        "email VARCHAR(250), "
                        "PRIMARY KEY (tipo_doc, nro_doc)"
                      ");")
            c.execute("CREATE TABLE domicilio ("
                        "id INTEGER PRIMARY KEY AUTOINCREMENT, "
                        "tipo_doc INTEGER, "
                        "nro_doc INTEGER, "
                        "direccion TEXT, "
                        "FOREIGN KEY (tipo_doc, nro_doc) REFERENCES padron "
                      ");")
            # importar los datos a la base sqlite
            for i, l in enumerate(f):
                if i % 10000 == 0:
                    print(i)
                l = l.strip("\x00")
                r = leer(l, FORMATO)
                params = [r[k] for k in keys]
                params[8] = 80          # agrego tipo_doc = CUIT
                params[9] = None        # cat_iva no viene de AFIP
                placeholders = ", ".join(["?"] * len(params))
                c.execute("INSERT INTO padron VALUES (%s)" % placeholders,
                          params)
            db.commit()
            c.close()
            db.close()

    @inicializar_y_capturar_excepciones_simple
    def Buscar(self, nro_doc, tipo_doc=80):
        "Devuelve True si fue encontrado y establece atributos con datos"
        # cuit: codigo único de identificación tributaria del contribuyente
        #       (sin guiones)
        self.cursor.execute("SELECT * FROM padron WHERE "
                            " tipo_doc=? AND nro_doc=?", [tipo_doc, nro_doc])
        row = self.cursor.fetchone()
        for key in [k for k, l, t, d in FORMATO]:
            if row:
                val = row[key]
                if not isinstance(val, str):
                    val = str(row[key])
                setattr(self, key, val)
            else:
                setattr(self, key, '')
        if self.tipo_doc == 80:
            self.cuit = self.nro_doc
        elif self.tipo_doc == 96:
            self.dni = self.nro_doc
        return True if row else False

    @inicializar_y_capturar_excepciones_simple
    def ConsultarDomicilios(self, nro_doc, tipo_doc=80, cat_iva=None):
        "Busca los domicilios, devuelve la cantidad y establece la lista"
        self.cursor.execute("SELECT direccion FROM domicilio WHERE "
                            " tipo_doc=? AND nro_doc=? ORDER BY id ",
                            [tipo_doc, nro_doc])
        filas = self.cursor.fetchall()
        self.domicilios = [fila['direccion'] for fila in filas]
        return len(filas)

    @inicializar_y_capturar_excepciones_simple
    def Guardar(
        self, tipo_doc, nro_doc, denominacion, cat_iva, direccion, email
    ):
        "Agregar o actualizar los datos del cliente"
        if self.Buscar(nro_doc, tipo_doc):
            sql = ("UPDATE padron SET denominacion=?, cat_iva=?, email=? "
                    "WHERE tipo_doc=? AND nro_doc=?")
            params = [denominacion, cat_iva, email, tipo_doc, nro_doc]
        else:
            sql = ("INSERT INTO padron (tipo_doc, nro_doc, denominacion, "
                    "cat_iva, email) VALUES (?, ?, ?, ?, ?)")
            params = [tipo_doc, nro_doc, denominacion, cat_iva, email]
        self.cursor.execute(sql, params)
        # agregar el domicilio solo si no existe:
        if direccion:
            self.cursor.execute("SELECT * FROM domicilio WHERE direccion=? "
                                "AND tipo_doc=? AND nro_doc=?",
                                [direccion, tipo_doc, nro_doc])
            if self.cursor.rowcount < 0:
                sql = ("INSERT INTO domicilio (nro_doc, tipo_doc, direccion)"
                        "VALUES (?, ?, ?)")
                self.cursor.execute(sql, [nro_doc, tipo_doc, direccion])
        self.db.commit()
        return True

    @inicializar_y_capturar_excepciones_simple
    def Consultar(self, nro_doc):
        "Llama a la API pública de AFIP para obtener los datos de una persona"
        if not self.client:
            self.Conectar()
        self.response = self.client("sr-padron", "v2", "persona", str(nro_doc))
        result = json.loads(self.response)
        if result['success']:
            data = result['data']
            # extraigo datos generales del contribuyente:
            self.cuit = data["idPersona"]
            self.tipo_persona = data["tipoPersona"]
            self.tipo_doc = TIPO_CLAVE.get(data["tipoClave"])
            self.dni = data.get("numeroDocumento")
            self.estado = data.get("estadoClave")
            self.denominacion = data.get("nombre")
            # analizo el domicilio
            domicilio = data.get("domicilioFiscal")
            if domicilio:
                self.direccion = domicilio.get("direccion", "")
                self.localidad = domicilio.get("localidad", "")  # no usado en
                # CABA
                self.provincia = PROVINCIAS.get(
                    domicilio.get("idProvincia"), ""
                )
                self.cod_postal = domicilio.get("codPostal")
            else:
                self.direccion = self.localidad = self.provincia = ""
                self.cod_postal = ""
            # retrocompatibilidad:
            self.domicilios = [
                "%s - %s (%s) - %s" % (
                    self.direccion, self.localidad, self.cod_postal,
                    self.provincia,)
                ]
            # analizo impuestos:
            self.impuestos = data.get("impuestos", [])
            self.actividades = data.get("actividades", [])
            self.imp_iva = "S" if 30 in self.impuestos else "N"
            mt = data.get("categoriasMonotributo", {})
            self.monotributo = "S" if mt else "N"
            self.actividad_monotributo = ""  # TODO: mt[0].get("idCategoria")
            self.integrante_soc = ""
            self.empleador = "S" if 301 in self.impuestos else "N"
            self.cat_iva = ""
            self.data = data
        else:
            error = result['error']
            self.Excepcion = error['mensaje']
        return True

    @inicializar_y_capturar_excepciones_simple
    def DescargarConstancia(self, nro_doc, filename="constancia.pdf"):
        "Llama a la API para descargar una constancia de inscripcion (PDF)"
        if not self.client:
            self.Conectar()
        self.response = self.client(
            "sr-padron", "v1", "constancia", str(nro_doc)
        )
        if self.response.startswith("{"):
            result = json.loads(self.response)
            assert not result["success"]
            self.Excepcion = result['error']['mensaje']
            return False
        else:
            with open(filename, "wb") as f:
                f.write(self.response)
            return True

    @inicializar_y_capturar_excepciones_simple
    def MostrarPDF(self, archivo, imprimir=False):
        if sys.platform.startswith(("linux2", 'java')):
            os.system("evince ""%s""" % archivo)
        else:
            operation = imprimir and "print" or ""
            os.startfile(archivo, operation)
        return True

    def ObtenerTablaParametros(self, tipo_recurso, sep="||"):
        "Devuelve un array de elementos que tienen id y descripción"
        if not self.client:
            self.Conectar()
        self.response = self.client("parametros", "v1", tipo_recurso)
        result = json.loads(self.response)
        ret = {}
        if result['success']:
            data = result['data']
            # armo un diccionario con los datos devueltos:
            key = [k for k in list(data[0].keys()) if k.startswith("id")][0]
            val = [k for k in list(data[0].keys()) if k.startswith("desc")][0]
            for it in data:
                ret[it[key]] = it[val]
            self.data = data
        else:
            error = result['error']
            self.Excepcion = error['mensaje']
        if sep:
            return ["%s%%s%s%%s%s" % (sep, sep, sep) % it for it in sorted(ret.items())]
        else:
            return ret



# busco el directorio de instalación (global para que no cambie si usan otra
# dll)
INSTALL_DIR = PadronAFIP.InstallDir = get_install_dir()

if __name__ == "__main__":

    safe_console()

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(PadronAFIP)
    else:
        padron = PadronAFIP()
        padron.LanzarExcepciones = True
        import time
        t0 = time.time()
        if "--descargar" in sys.argv:
            padron.Descargar()
        if "--procesar" in sys.argv:
            padron.Procesar(borrar='--borrar' in sys.argv)
        cuit = len(sys.argv) > 1 and sys.argv[1] or "20267565393"
        # consultar un cuit:
        if '--online' in sys.argv:
            padron.Conectar(trace="--trace" in sys.argv)
            print("Consultando AFIP online...")
            ok = padron.Consultar(cuit)
            print(('ok' if ok else "error", padron.Excepcion))
            print(("Denominacion:", padron.denominacion))
            print(("CUIT:", padron.cuit))
            print(("Tipo:", padron.tipo_persona, padron.tipo_doc, padron.dni))
            print(("Estado:", padron.estado))
            print(("Direccion:", padron.direccion))
            print(("Localidad:", padron.localidad))
            print(("Provincia:", padron.provincia))
            print(("Codigo Postal:", padron.cod_postal))
            print(("Impuestos:", padron.impuestos))
            print(("Actividades:", padron.actividades))
            print(("IVA", padron.imp_iva))
            print(("MT", padron.monotributo, padron.actividad_monotributo))
            print(("Empleador", padron.empleador))
        elif '--constancia' in sys.argv:
            filename = sys.argv[2]
            print(("Descargando constancia AFIP online...", cuit, filename))
            ok = padron.DescargarConstancia(cuit, filename)
            print(('ok' if ok else "error", padron.Excepcion))
            if '--mostrar' in sys.argv:
                padron.MostrarPDF(archivo=filename,
                                 imprimir='--imprimir' in sys.argv)
        else:
            ok = padron.Buscar(cuit)
            if ok:
                print(("Denominacion:", padron.denominacion))
                print(("IVA:", padron.imp_iva))
                padron.ConsultarDomicilios(cuit)
                for dom in padron.domicilios:
                    print(dom)
            else:
                print((padron.Excepcion))
                print((padron.Traceback))
        t1 = time.time()
        if '--trace' in sys.argv:
            print(("tiempo", t1 - t0))

